#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
swap.proto.thorchain.thornode: THORChain swap protocol network query ops
"""

import time, json
from collections import namedtuple

from ....protocol import init_proto
from ....amt import UniAmt
from ....http import HTTPClient

_gd = namedtuple('gas_unit_data', ['code', 'disp'])
gas_unit_data = {
	'satsperbyte': _gd('s', 'sat/byte'),
	'gwei':        _gd('G', 'Gwei'),
}

class ThornodeSwapClient(HTTPClient):

	http_hdrs = {'Content-Type': 'application/json'}
	timeout = 5

	def __init__(self, tx, *, network_proto=None, host=None):
		rune_proto = init_proto(tx.cfg, 'rune', network=tx.cfg._proto.network)
		for k, v in rune_proto.rpc_swap_params.items():
			setattr(self, k, v)
		super().__init__(tx.cfg, network_proto=network_proto, host=host)

class Thornode:

	def __init__(self, tx, amt):
		self.tx = tx
		self.in_amt = UniAmt(f'{amt:.8f}')
		self.rpc = ThornodeSwapClient(tx)

	def get_quote(self, swap_cfg):

		def get_data(send, recv, amt):
			get_str = (
				'/thorchain/quote/swap?'
				f'from_asset={send}&'
				f'to_asset={recv}&'
				f'amount={amt}&'
				f'streaming_interval={swap_cfg.stream_interval}')
			data = json.loads(self.rpc.get(path=get_str))
			if not 'expiry' in data:
				from ....util import pp_fmt, die
				die(2, pp_fmt(data))
			return data

		if (
				(self.tx.proto.tokensym or self.tx.recv_asset.tokensym)
				and not self.tx.send_asset.chain == 'THOR'): # token swap
			in_data = get_data(
				self.tx.send_asset.full_name,
				'THOR.RUNE',
				self.in_amt.to_unit('satoshi'))
			if self.tx.proto.network != 'regtest':
				time.sleep(1.1) # ninerealms max request rate 1/sec
			out_data = get_data(
				'THOR.RUNE',
				self.tx.recv_asset.full_name,
				in_data['expected_amount_out'])
			self.data = in_data | {
				'expected_amount_out': out_data['expected_amount_out'],
				'fees': out_data['fees'],
				'expiry': min(in_data['expiry'], out_data['expiry'])
			}
		else:
			self.data = get_data(
				self.tx.send_asset.full_name,
				self.tx.recv_asset.full_name,
				self.in_amt.to_unit('satoshi'))

	async def format_quote(self, trade_limit, *, deduct_est_fee=False):
		from ....util import make_timestr, ymsg
		from ....util2 import format_elapsed_hr
		from ....color import blue, green, cyan, pink, orange, redbg, yelbg, grnbg
		from . import name

		d = self.data
		tx = self.tx
		in_coin = tx.send_asset.short_name
		out_coin = tx.recv_asset.short_name
		in_amt = self.in_amt
		out_amt = UniAmt(int(d['expected_amount_out']), from_unit='satoshi')
		if tx.proto.has_usr_fee:
			gas_unit = d['gas_rate_units']

		if trade_limit:
			from . import ExpInt4
			tl_int = ExpInt4(trade_limit.to_unit('satoshi'))
			tl_uniamt = UniAmt(tl_int.trunc, from_unit='satoshi')
			ratio = float(tl_uniamt / out_amt)
			direction = 'ABOVE' if ratio > 1 else 'below'
			mcolor, lblcolor = (
				(redbg, redbg) if (ratio < 0.93 or ratio > 0.999) else
				(yelbg, yelbg) if ratio < 0.97 else
				(green, grnbg))
			trade_limit_disp = f"""
  {lblcolor('Trade limit:')}                   {tl_uniamt.hl()} {out_coin} """ + mcolor(
				f'({abs(1 - ratio) * 100:0.2f}% {direction} expected amount)')
			tx_size_adj = len(tl_int.enc) - 1
			if tx.proto.is_evm:
				tx.adj_gas_with_extra_data_len(len(tl_int.enc) - 1) # one-shot method, no-op if repeated
		else:
			trade_limit_disp = ''
			tx_size_adj = 0

		def get_estimated_fee():
			return tx.feespec2abs(
				fee_arg = d['recommended_gas_rate'] + gas_unit_data[gas_unit].code,
				tx_size = None if tx.proto.is_vm else tx.estimate_size() + tx_size_adj)

		_amount_in_label = 'Amount in:'
		if deduct_est_fee:
			if not tx.proto.has_usr_fee:
				in_amt -= tx.usr_fee
				out_amt *= (in_amt / self.in_amt)
				_amount_in_label = 'Amount in:'
			elif gas_unit in gas_unit_data:
				in_amt -= UniAmt(f'{get_estimated_fee():.8f}')
				out_amt *= (in_amt / self.in_amt)
				_amount_in_label = 'Amount in (estimated):'
			else:
				ymsg(f'Warning: unknown gas unit ‘{gas_unit}’, cannot estimate fee')

		min_in_amt = UniAmt(int(d['recommended_min_amount_in']), from_unit='satoshi')
		elapsed_disp = format_elapsed_hr(d['expiry'], future_msg='from now')
		fees = d['fees']
		fees_t = UniAmt(int(fees['total']), from_unit='satoshi')
		fees_pct_disp = str(fees['total_bps'] / 100) + '%'
		slip_pct_disp = str(fees['slippage_bps'] / 100) + '%'
		hdr = f'SWAP QUOTE (source: {self.rpc.host})'

		vault_info = '' if tx.send_asset.chain == 'THOR' else """
  Vault address:                 {}""".format(cyan(self.inbound_address))

		fee_info = '' if not tx.proto.has_usr_fee else """
  Recommended fee:               {} {}
  Network-estimated fee:         {} (from node)""".format(
			pink(d['recommended_gas_rate']),
			pink(_.disp if (_ := gas_unit_data.get(gas_unit)) else gas_unit),
			await self.tx.network_fee_disp())

		return f"""
{cyan(hdr)}
  Protocol:                      {blue(name)}
  Direction:                     {orange(f'{tx.send_asset.name} => {tx.recv_asset.name}')}{vault_info}
  Quote expires:                 {pink(elapsed_disp)} [{make_timestr(d['expiry'])}]
  {_amount_in_label:<22}         {in_amt.hl()} {in_coin}
  Expected amount out:           {out_amt.hl()} {out_coin}{trade_limit_disp}
  Rate:                          {(out_amt / in_amt).hl()} {out_coin}/{in_coin}
  Reverse rate:                  {(in_amt / out_amt).hl()} {in_coin}/{out_coin}
  Recommended minimum in amount: {min_in_amt.hl()} {in_coin}{fee_info}
  Fees:
    Total:    {fees_t.hl()} {out_coin} ({pink(fees_pct_disp)})
    Slippage: {pink(slip_pct_disp)}
"""

	@property
	def inbound_address(self):
		addr = self.data['inbound_address']
		return addr.removeprefix('0x') if self.tx.proto.is_evm else addr

	@property
	def router(self):
		return self.data['router'].lower().removeprefix('0x')

	@property
	def rel_fee_hint(self):
		if self.tx.proto.has_usr_fee:
			gas_unit = self.data['gas_rate_units']
			if gas_unit in gas_unit_data:
				return self.data['recommended_gas_rate'] + gas_unit_data[gas_unit].code

	def __str__(self):
		from pprint import pformat
		return pformat(self.data)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
wallet.plainhex: plain hexadecimal wallet class
"""

from ..util import msg, is_hex_str_lc
from ..seed import Seed
from .unenc import wallet

class wallet(wallet):

	stdin_ok = True
	desc = 'plain hexadecimal seed data'

	def _format(self):
		self.fmt_data = self.seed.data.hex() + '\n'

	def _deformat(self):
		desc = self.desc
		d = self.fmt_data.strip()

		if not is_hex_str_lc(d):
			msg(f'{d!r}: not a lowercase hexadecimal string, in {desc}')
			return False

		if not len(d)*4 in Seed.lens:
			msg(f'Invalid data length ({len(d)}) in {desc}')
			return False

		self.seed = Seed(self.cfg, seed_bin=bytes.fromhex(d))

		self.check_usr_seed_len()

		return True



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>

"""
test.objtest_d.ltc_testnet: LTC testnet test vectors for MMGen data objects
"""

from mmgen.addr import CoinAddr
from mmgen.key import WifKey, PrivKey
from mmgen.protocol import init_proto

from .common import r16, r32
from ..include.common import cfg

proto = init_proto(cfg, 'ltc', network='testnet', need_amt=True)

tests = {
	'CoinAddr': {
		'bad': (
			{'addr': 1,   'proto': proto},
			{'addr': 'x', 'proto': proto},
			{'addr': 'я', 'proto': proto},
		),
		'good': (
			{'addr': 'n2D3joAy3yE5fqxUeCp38X6uPUcVn7EFw9', 'proto': proto},
			{'addr': 'QN59YbnHsPQcbKWSq9PmTpjrhBnHGQqRmf', 'proto': proto},
		),
	},
	'WifKey': {
		'bad': (
			{'proto': proto, 'wif': 1},
			{'proto': proto, 'wif': []},
			{'proto': proto, 'wif': '\0'},
			{'proto': proto, 'wif': '\1'},
			{'proto': proto, 'wif': 'я'},
			{'proto': proto, 'wif': 'g'},
			{'proto': proto, 'wif': 'gg'},
			{'proto': proto, 'wif': 'FF'},
			{'proto': proto, 'wif': 'f00'},
			{'proto': proto, 'wif': r16.hex()},
			{'proto': proto, 'wif': '2MspvWFjBbkv2wzQGqhxJUYPCk3Y2jMaxLN'},
		),
		'good': (
			{'proto': proto, 'wif': '936Fd4qs3Zy2ZiYHH7vZ3UpT23KtCAiGiG2xBTkjHo7jE9aWA2f',  'ret_idx': 1},
			{'proto': proto, 'wif': 'cQY3EumdaSNuttvDSUuPdiMYLyw8aVmYfFqxo9kdPuWbJBN4Ny66', 'ret_idx': 1},
		)
	},
	'PrivKey': {
		'bad': (
			{'proto': proto, 'wif': 1},
			{'proto': proto, 'wif': '1'},
			{'proto': proto, 'wif': '5KXEpVzjWreTcQoG5hX357s1969MUKNLuSfcszF6yu84kpsNZKb'},
			{'proto': proto, 's': r32, 'wif': '5KXEpVzjWreTcQoG5hX357s1969MUKNLuSfcszF6yu84kpsNZKb'},
			{'proto': proto, 'pubkey_type': 'std'},
			{'proto': proto, 's': r32},
			{'proto': proto, 's': r32, 'compressed': 'yes'},
			{'proto': proto, 's': r32, 'compressed': 'yes', 'pubkey_type': 'std'},
			{'proto': proto, 's': r32, 'compressed': True, 'pubkey_type': 'nonstd'},
			{'proto': proto, 's': r32+b'x', 'compressed': True, 'pubkey_type': 'std'}
		),
		'good': (
			{'proto': proto, 'wif': '92iqzh6NqiKawyB1ronw66YtEHrU4rxRJ5T4aHniZqvuSVZS21f',
			'ret': bytes.fromhex('95b2aa7912550eacdd3844dcc14bee08ce7bc2434ad4858beb136021e945afeb')},
			{'proto': proto, 'wif': 'cSaJAXBAm9ooHpVJgoxqjDG3AcareFy29Cz8mhnNTRijjv2HLgta',
			'ret': bytes.fromhex('94fa8b90c11fea8fb907c9376b919534b0a75b9a9621edf71a78753544b4101c')},
			{'proto': proto, 's': r32, 'compressed': False, 'pubkey_type': 'std', 'ret': r32},
			{'proto': proto, 's': r32, 'compressed': True, 'pubkey_type': 'std', 'ret': r32}
		)
	},
}



from .asset_orig import *

class overlay_fake_THORChainSwapAsset:

	assets_data = {
		'ETH.MM1':  THORChainSwapAsset._ad('MM1 Token (ETH)',  None, 'ETH.MM1',  None, True),
		'ETH.JUNK': THORChainSwapAsset._ad('Junk Token (ETH)', None, 'ETH.JUNK', None, True),
		'ETH.NONE': THORChainSwapAsset._ad('Unavailable Token (ETH)', None, 'ETH.NONE', None, True)
	}
	evm_contracts = {
		'ETH.MM1':  'deadbeefdeadbeefdeadbeefdeadbeefdeadbeef'
	}
	unsupported = ('ETH.NONE',)

	blacklisted = {
		'ETH.JUNK': 'Because it’s junk',
	}

THORChainSwapAsset.assets_data |= overlay_fake_THORChainSwapAsset.assets_data
THORChainSwapAsset.unsupported += overlay_fake_THORChainSwapAsset.unsupported
THORChainSwapAsset.blacklisted.update(overlay_fake_THORChainSwapAsset.blacklisted)
THORChainSwapAsset.evm_contracts.update(overlay_fake_THORChainSwapAsset.evm_contracts)



import os as overlay_fake_os
from .new_orig import *

if overlay_fake_os.getenv('MMGEN_BOGUS_UNSPENT_DATA'):

	class overlay_fake_New(New):

		async def warn_addr_used(self, proto, chg, desc):
			from ..util import ymsg
			ymsg('Bogus unspent data: skipping used change address check')

	New.warn_addr_used = overlay_fake_New.warn_addr_used



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
test.cmdtest_d.automount_eth: Ethereum automount autosigning tests for the cmdtest.py test suite
"""
import os, re

from .autosign import CmdTestAutosignThreaded
from .ethdev import CmdTestEthdev, CmdTestEthdevMethods
from .include.common import dfl_words_file
from ..include.common import cfg, joinpath

class CmdTestAutosignETH(CmdTestAutosignThreaded, CmdTestEthdev, CmdTestEthdevMethods):
	'automounted transacting operations for Ethereum via ethdev'

	networks = ('eth', 'etc')
	tmpdir_nums = [59]

	cmd_group = (
		('setup',                  f'dev mode tests for coin {cfg.coin} (start daemon)'),
		('addrgen',                'generating addresses'),
		('addrimport',             'importing addresses'),
		('addrimport_devaddr',     'importing the dev address'),
		('addrimport_reth_devaddr','importing the reth dev address'),
		('fund_devaddr',           'funding the dev address'),
		('del_reth_devaddr',       'deleting the reth dev address'),
		('fund_mmgen_addr',        'funding an MMGen address'),
		('create_tx',              'creating a transaction'),
		('run_autosign_setup',     'running ‘autosign setup’'),
		('wait_loop_start',        'starting autosign wait loop'),
		('send_tx',                'sending the transaction'),
		('token_compile1',         'compiling ERC20 token #1'),
		('token_deploy1a',         'deploying ERC20 token #1 (SafeMath)'),
		('token_deploy1b',         'deploying ERC20 token #1 (Owned)'),
		('token_deploy1c',         'deploying ERC20 token #1 (Token)'),
		('tx_status2',             'getting the transaction status'),
		('token_fund_user',        'transferring token funds from dev to user'),
		('token_addrgen',          'generating token addresses'),
		('token_addrimport',       'importing token addresses using token address (MM1)'),
		('token_bal1',             f'the {cfg.coin} balance and token balance'),
		('create_token_tx',        'creating a token transaction'),
		('send_token_tx',          'sending a token transaction'),
		('token_bal2',             f'the {cfg.coin} balance and token balance'),
		('wait_loop_kill',         'stopping autosign wait loop'),
		('stop',                   'stopping daemon'),
		('txview',                 'viewing transactions'),
	)

	def __init__(self, cfg, trunner, cfgs, spawn):

		self.coins = [cfg.coin.lower()]

		CmdTestAutosignThreaded.__init__(self, cfg, trunner, cfgs, spawn)
		CmdTestEthdev.__init__(self, cfg, trunner, cfgs, spawn)

		self.txop_opts = ['--autosign', '--regtest=1', '--quiet']

	def fund_mmgen_addr(self):
		return self._fund_mmgen_addr(arg='98831F3A:E:1,123.456')

	def create_tx(self):
		self.mining_delay()
		self.insert_device_online()
		t = self._create_tx(fee='50G', args=['98831F3A:E:11,54.321'], add_opts=self.txop_opts)
		t.read()
		self.remove_device_online()
		return t

	def run_autosign_setup(self):
		return self.run_setup(mn_type='bip39', mn_file='test/ref/98831F3A.bip39', use_dfl_wallet=None)

	def send_tx(self):
		self._wait_signed('transaction')
		self.insert_device_online()
		t = self._send_tx(desc='automount transaction', add_opts=self.txop_opts)
		t.read()
		self.remove_device_online()
		return t

	def token_addrgen(self):
		return self._token_addrgen(mm_idxs=[11], naddrs=3)

	def token_addrimport(self):
		return self._token_addrimport('token_addr1', '11-13', expect='3/3')

	def token_fund_user(self):
		return self._token_transfer_ops(op='fund_user', mm_idxs=[11])

	def token_bal1(self):
		return self._token_bal_check(pat=r':E:11\s+1000\s+54\.321\s+')

	def token_bal2(self):
		return self._token_bal_check(pat=r':E:11\s+998.76544\s+54.318\d+\s+.*:E:12\s+1\.23456\s+')

	def create_token_tx(self):
		self.insert_device_online()
		t = self._create_token_tx(
			cmd = 'txcreate',
			fee = '50G',
			args = ['98831F3A:E:12,1.23456'],
			add_opts = self.txop_opts)
		t.read()
		self.remove_device_online()
		return t

	send_token_tx = send_tx



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
help.seedsplit: seedsplit help notes for MMGen suite
"""

def help(proto, cfg):
	from ..seedsplit import SeedShareIdx, SeedShareCount, MasterShareIdx
	return """
COMMAND NOTES:

This command generates shares one at a time.  Shares may be output to any
MMGen wallet format, with one limitation: only one share in a given split may
be in hidden incognito format, and it must be the master share in the case of
a master-share split.

If the command's optional first argument is omitted, the default wallet is
used for the split.

The last argument is a seed split specifier consisting of an optional split
ID, a share index, and a share count, all separated by colons.  The split ID
must be a valid UTF-8 string.  If omitted, the ID 'default' is used.  The
share index (the index of the share being generated) must be in the range
{si.min_val}-{si.max_val} and the share count (the total number of shares in the split)
in the range {sc.min_val}-{sc.max_val}.

Master Shares

Each seed has a total of {mi.max_val} master shares, which can be used as the first
shares in multiple splits if desired.  To generate a master share, use the
--master-share (-M) option with an index in the range {mi.min_val}-{mi.max_val} and omit
the last argument.

When creating and joining a split using a master share, ensure that the same
master share index is used in all split and join commands.

EXAMPLES:

  Split a BIP39 seed phrase into two BIP39 shares.  Rejoin the split:

    $ echo 'zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo wrong' > sample.bip39

    $ mmgen-seedsplit -o bip39 sample.bip39 1:2
    BIP39 mnemonic data written to file '03BAE887-default-1of2[D51CB683][128].bip39'

    $ mmgen-seedsplit -o bip39 sample.bip39 2:2
    BIP39 mnemonic data written to file '03BAE887-default-2of2[67BFD36E][128].bip39'

    $ mmgen-seedjoin -o bip39 \\
        '03BAE887-default-2of2[67BFD36E][128].bip39' \\
        '03BAE887-default-1of2[D51CB683][128].bip39'
    BIP39 mnemonic data written to file '03BAE887[128].bip39'

    $ cat '03BAE887[128].bip39'
    zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo zoo wrong

  Create a 3-way default split of your default wallet, outputting all shares
  to default wallet format.  Rejoin the split:

    $ mmgen-seedsplit 1:3 # Step A
    $ mmgen-seedsplit 2:3 # Step B
    $ mmgen-seedsplit 3:3 # Step C
    $ mmgen-seedjoin <output_of_step_A> <output_of_step_B> <output_of_step_C>

  Create a 2-way split of your default wallet with ID string 'alice',
  outputting shares to MMGen native mnemonic format.  Rejoin the split:

    $ mmgen-seedsplit -o words alice:1:2 # Step D
    $ mmgen-seedsplit -o words alice:2:2 # Step E
    $ mmgen-seedjoin <output_of_step_D> <output_of_step_E>

  Create a 2-way split of your default wallet with ID string 'bob' using
  master share #7, outputting share #1 (the master share) to default wallet
  format and share #2 to BIP39 format.  Rejoin the split:

    $ mmgen-seedsplit -M7                   # Step X
    $ mmgen-seedsplit -M7 -o bip39 bob:2:2  # Step Y
    $ mmgen-seedjoin -M7 --id-str=bob <output_of_step_X> <output_of_step_Y>

  Create a 2-way split of your default wallet with ID string 'alice' using
  master share #7.  Rejoin the split using master share #7 generated in the
  previous example:

    $ mmgen-seedsplit -M7 -o bip39 alice:2:2 # Step Z
    $ mmgen-seedjoin -M7 --id-str=alice <output_of_step_X> <output_of_step_Z>

  Create a 2-way default split of your default wallet with an incognito-format
  master share hidden in file 'my.hincog' at offset 1325.  Rejoin the split:

    $ mmgen-seedsplit -M4 -o hincog -J my.hincog,1325 1:2 # Step M (share A)
    $ mmgen-seedsplit -M4 -o bip39 2:2                    # Step N (share B)
    $ mmgen-seedjoin -M4 -H my.hincog,1325 <output_of_step_N>

""".strip().format(
	si = SeedShareIdx,
	sc = SeedShareCount,
	mi = MasterShareIdx)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.xchain.addrgen: Cross-chain address generation classes for the MMGen suite
"""

from ...addrgen import addr_generator, check_data
from ..btc.common import hash160

class bech32x(addr_generator.base):

	@check_data
	def to_addr(self, data):
		return self.proto.encode_addr_bech32x(hash160(data.pubkey))



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
tx.base: base transaction class
"""

from ..cfg import gc
from ..objmethods import MMGenObject
from ..obj import (
	ImmutableAttr,
	ListItemAttr,
	MMGenListItem,
	MMGenTxComment,
	TwComment,
	CoinTxID,
	HexStr,
	NonNegativeInt
)
from ..amt import CoinAmtChk
from ..addr import MMGenID, CoinAddr
from ..util import msg, ymsg, fmt, remove_dups, make_timestamp, die, cached_property

class MMGenTxIO(MMGenListItem):
	vout     = ListItemAttr(NonNegativeInt)
	amt      = ImmutableAttr(CoinAmtChk, include_proto=True)
	comment  = ListItemAttr(TwComment, reassign_ok=True)
	mmid     = ListItemAttr(MMGenID, include_proto=True)
	addr     = ImmutableAttr(CoinAddr, include_proto=True)
	confs    = ListItemAttr(int) # confs of type long exist in the wild, so convert
	txid     = ListItemAttr(CoinTxID)
	have_wif = ListItemAttr(bool, typeconv=False, delete_ok=True)

	invalid_attrs = {'proto', 'tw_copy_attrs'}

	def __init__(self, proto, **kwargs):
		self.__dict__['proto'] = proto
		MMGenListItem.__init__(self, **kwargs)

	@property
	def mmtype(self):
		"""
		Attempt to determine input or output’s MMGenAddrType.  For non-MMGen
		addresses, infer the type from the address format, returning None for
		P2PKH, which could be either 'L' or 'C'.
		"""
		return (
			str(self.mmid.mmtype) if self.mmid else
			'B' if self.addr.addr_fmt == 'bech32' else
			'S' if self.addr.addr_fmt == 'p2sh' else
			None
		) if self.addr else None

class MMGenTxIOList(list, MMGenObject):

	def __init__(self, parent, data=None):
		self.parent = parent
		if data:
			assert isinstance(data, list), 'MMGenTxIOList_check1'
		else:
			data = []
		list.__init__(self, data)

class Base(MMGenObject):
	desc         = 'transaction'
	comment      = None
	txid         = None
	coin_txid    = None
	timestamp    = None
	sent_timestamp = None
	blockcount   = None
	locktime     = None
	chain        = None
	signed       = False
	is_bump      = False
	is_swap      = False
	swap_attrs = {
		'swap_proto': None,
		'swap_quote_expiry': None,
		'swap_recv_addr_mmid': None,
		'swap_recv_asset_spec': None,
		'swap_memo': None,
		'token_vault_addr': None,
		'serialized2': None,
		'coin_txid2': CoinTxID}
	file_format  = 'json'
	non_mmgen_inputs_msg = f"""
		This transaction includes inputs with non-{gc.proj_name} addresses.  When
		signing the transaction, private keys for the addresses listed below must
		be supplied using the --keys-from-file option.  The key file must contain
		one key per line.

		Non-{gc.proj_name} addresses found in inputs:
		  {{}}
	"""

	class Input(MMGenTxIO):
		scriptPubKey  = ListItemAttr(HexStr)
		sequence      = ListItemAttr(int, typeconv=False)
		tw_copy_attrs = {'scriptPubKey', 'vout', 'amt', 'comment', 'mmid', 'addr', 'confs', 'txid'}

	class Output(MMGenTxIO):
		addr     = ListItemAttr(CoinAddr, include_proto=True) # ImmutableAttr in parent cls
		is_chg   = ListItemAttr(bool, typeconv=False)
		is_vault = ListItemAttr(bool, typeconv=False)
		data     = ListItemAttr(None, typeconv=False) # placeholder

	class InputList(MMGenTxIOList):
		desc = 'transaction inputs'

	class OutputList(MMGenTxIOList):
		desc = 'transaction outputs'

	def __init__(self, *args, **kwargs):
		self.cfg      = kwargs['cfg']
		self.inputs   = self.InputList(self)
		self.outputs  = self.OutputList(self)
		self.name     = type(self).__name__
		self.proto    = kwargs['proto']
		self.twctl    = kwargs.get('twctl')
		self.is_token = 'Token' in self.name

	@property
	def coin(self):
		return self.proto.coin

	@property
	def dcoin(self):
		return self.proto.dcoin

	@property
	def info(self):
		from .info import init_info
		return init_info(self.cfg, self)

	def check_correct_chain(self):
		if hasattr(self, 'rpc'):
			if self.chain != self.rpc.chain:
				die('TransactionChainMismatch',
					f'Transaction is for {self.chain}, but coin daemon chain is {self.rpc.chain}!')

	def sum_inputs(self):
		return sum(e.amt for e in self.inputs)

	def sum_outputs(self, *, exclude=None):
		if exclude is None:
			olist = self.outputs
		else:
			olist = self.outputs[:exclude] + self.outputs[exclude+1:]
		if not olist:
			return self.proto.coin_amt('0')
		return sum(e.amt for e in olist)

	def _chg_output_ops(self, op, attr):
		is_chgs = [getattr(x, attr) for x in self.outputs]
		if is_chgs.count(True) == 1:
			return is_chgs.index(True) if op == 'idx' else self.outputs[is_chgs.index(True)]
		elif is_chgs.count(True) == 0:
			return None
		else:
			raise ValueError('more than one change output!')

	@property
	def chg_idx(self):
		return self._chg_output_ops('idx', 'is_chg')

	@property
	def chg_output(self):
		return self._chg_output_ops('output', 'is_chg')

	def add_timestamp(self):
		self.timestamp = make_timestamp()

	def add_sent_timestamp(self):
		self.sent_timestamp = make_timestamp()

	def add_blockcount(self):
		self.blockcount = self.rpc.blockcount

	# returns True if comment added or changed, False otherwise
	def add_comment(self, *, infile=None):
		if infile:
			from ..fileutil import get_data_from_file
			self.comment = MMGenTxComment(
				get_data_from_file(self.cfg, infile, desc='transaction comment'))
		else:
			from ..ui import keypress_confirm, line_input
			if keypress_confirm(
					self.cfg,
					prompt = 'Edit transaction comment?' if self.comment else 'Add a comment to transaction?',
					default_yes = False):
				res = MMGenTxComment(line_input(self.cfg, 'Comment: ', insert_txt=self.comment))
				if not res:
					ymsg('Warning: comment is empty')
				changed = res != self.comment
				self.comment = res
				return changed
			else:
				return False

	def get_non_mmaddrs(self, desc):
		return remove_dups(
			(i.addr for i in getattr(self, desc) if not i.mmid),
			edesc = 'non-MMGen address',
			quiet = True)

	def check_non_mmgen_inputs(self, *, caller, non_mmaddrs=None):
		assert caller in ('txcreate', 'txdo', 'txsign', 'autosign')
		non_mmaddrs = non_mmaddrs or self.get_non_mmaddrs('inputs')
		if non_mmaddrs:
			indent = '  '
			fs = fmt(self.non_mmgen_inputs_msg, strip_char='\t', indent=indent).strip()
			m = fs.format('\n    '.join(non_mmaddrs))
			if caller in ('txdo', 'txsign'):
				if not self.cfg.keys_from_file:
					die('UserOptError', f'\n{indent}ERROR: {m}\n')
			else:
				msg(f'\n{indent}WARNING: {m}\n')
				if not (caller == 'autosign' or self.cfg.yes):
					from ..ui import keypress_confirm
					keypress_confirm(self.cfg, 'Continue?', default_yes=True, do_exit=True)

	# swap methods:
	@cached_property
	def swap_proto_mod(self):
		from .new_swap import get_swap_proto_mod
		return get_swap_proto_mod(self.swap_proto)

	@cached_property
	def send_asset(self):
		spec = self.proto.coin + (f'.{self.proto.tokensym}' if self.proto.tokensym else '')
		return self.swap_proto_mod.SwapAsset(spec, 'send')

	@cached_property
	def recv_asset(self):
		if hasattr(self, 'swap_recv_asset_spec'):
			return self.swap_proto_mod.SwapAsset(self.swap_recv_asset_spec, 'recv')
		else: # backwards-compatibility workaround
			from ..swap.asset import SwapAsset
			x = '[unknown]'
			return SwapAsset._ad(x, x, x, x, x)

	# token methods:
	@property
	def token_op(self):
		return 'approve' if self.is_swap else 'transfer'



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
help.subwallet: subwallet help notes for the MMGen Wallet suite
"""

def help(proto, cfg):
	from ..subseed import SubSeedIdxRange
	return f"""
SUBWALLETS:

Subwallets (subseeds) are specified by a ‘Subseed Index’ consisting of:

  a) an integer in the range 1-{SubSeedIdxRange.max_idx}, plus
  b) an optional single letter, ‘L’ or ‘S’

The letter designates the length of the subseed.  If omitted, ‘L’ is assumed.

Long (‘L’) subseeds are the same length as their parent wallet’s seed
(typically 256 bits), while short (‘S’) subseeds are always 128-bit.
The long and short subseeds for a given index are derived independently,
so both may be used.

MMGen Wallet has no notion of ‘depth’, and to an outside observer subwallets
are identical to ordinary wallets.  This is a feature rather than a bug, as
it denies an attacker any way of knowing whether a given wallet has a parent.

Since subwallets are just wallets, they may be used to generate other
subwallets, leading to hierarchies of arbitrary depth.  However, this is
inadvisable in practice for two reasons:  Firstly, it creates accounting
complexity, requiring the user to independently keep track of a derivation
tree.  More importantly, however, it leads to the danger of Seed ID
collisions between subseeds at different levels of the hierarchy, as
MMGen checks and avoids ID collisions only among sibling subseeds.

An exception to this caveat would be a multi-user setup where sibling
subwallets are distributed to different users as their default wallets.
Since the subseeds derived from these subwallets are private to each user,
Seed ID collisions among them doesn’t present a problem.

A safe rule of thumb, therefore, is for *each user* to derive all of his/her
subwallets from a single parent.  This leaves each user with a total of two
million subwallets, which should be enough for most practical purposes.
""".strip()



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
test.cmdtest_d.include.proxy: SSH SOCKS proxy runner for the cmdtest.py test suite
"""

import sys, atexit
from subprocess import run, PIPE

from mmgen.util import msg, die, fmt
from mmgen.util2 import port_in_use

from ...include.common import omsg

class TestProxy:

	port = 49237
	no_ssh_errmsg = """
		The SSH daemon must be running and listening on localhost in order to test
		XMR TX relaying via SOCKS proxy.  If sshd is not running, please start it.
		Otherwise, add the line 'ListenAddress 127.0.0.1' to your sshd_config, and
		then restart the daemon.
	"""
	bad_perm_errmsg = """
		In order to test XMR TX relaying via SOCKS proxy, it’s desirable to enable
		SSH to localhost without a password, which is not currently supported by
		your configuration.  Your possible courses of action:

		1. Continue by answering 'y' at this prompt, and enter your system password
		   at the following prompt;

		2. Exit the test here, add your user SSH public key to your user
		   'authorized_keys' file, and restart the test; or

		3. Exit the test here, start the SSH SOCKS proxy manually by entering the
		   following command, and restart the test:

			  {}
	"""
	need_start_errmsg = """
		Please start the SSH SOCKS proxy by entering the following command:

			{}

		Then restart the test.
	"""

	def kill_proxy(self, args):
		if sys.platform in ('linux', 'darwin'):
			omsg(f'Stopping SSH SOCKS server at localhost:{self.port}')
			cmd = ['pkill', '-f', ' '.join(args)]
			run(cmd)

	def __init__(self, test_group, cfg):

		if test_group and test_group.is_helper:
			return

		def start_proxy():
			run(a + b2)
			omsg(f'SSH SOCKS server started, listening at localhost:{self.port}')

		a = ['ssh', '-x', '-o', 'ExitOnForwardFailure=True', '-D', f'localhost:{self.port}']
		b0 = ['-o', 'PasswordAuthentication=False']
		b1 = ['localhost', 'true']
		b2 = ['-fN', '-E', 'txrelay-proxy.debug', 'localhost']

		if port_in_use(self.port):
			omsg(f'Port {self.port} already in use. Assuming SSH SOCKS server is running')
		else:
			cp = run(a + b0 + b1, stdout=PIPE, stderr=PIPE)
			if err := cp.stderr.decode():
				omsg(err)
			if cp.returncode == 0:
				start_proxy()
			elif 'onnection refused' in err:
				die(2, fmt(self.no_ssh_errmsg, indent='    '))
			elif 'ermission denied' in err:
				msg(fmt(self.bad_perm_errmsg.format(' '.join(a + b2)), indent='    ', strip_char='\t'))
				from mmgen.ui import keypress_confirm
				keypress_confirm(cfg, 'Continue?', do_exit=True)
				start_proxy()
			else:
				die(2, fmt(self.need_start_errmsg.format(' '.join(a + b2)), indent='    '))

		if test_group is None:
			self.kill_proxy(a + b2)
		elif not hasattr(test_group.tr, 'proxy_stop_registered'):
			atexit.unregister(self.kill_proxy)
			atexit.register(self.kill_proxy, a + b2)
			test_group.tr.proxy_stop_registered = True



import os as overlay_fake_os
from .util2_orig import *

if overlay_fake_os.getenv('MMGEN_TEST_SUITE_DETERMINISTIC'):

	class overlay_fake_data:
		format_elapsed_hr = format_elapsed_hr
		format_elapsed_days_hr = format_elapsed_days_hr

	def format_elapsed_hr(t, now=None, cached={}):
		return overlay_fake_data.format_elapsed_hr(t, now=2204622671, cached=cached) # 11 Nov 2039 11:11:11

	def format_elapsed_days_hr(t, now=None, cached={}):
		return overlay_fake_data.format_elapsed_days_hr(t, now=2204622671, cached=cached)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
test.cmdtest_d.base: Base class for the cmdtest.py test suite
"""

import sys, os

from mmgen.util import msg
from mmgen.color import gray, purple, yellow

from ..include.common import write_to_file, read_from_file, imsg
from .include.common import get_file_with_ext

class CmdTestBase:
	'initializer class for the cmdtest.py test suite'
	base_passthru_opts = ('data_dir', 'skip_cfg_file')
	passthru_opts = ()
	networks = ()
	segwit_opts_ok = False
	color = False
	need_daemon = False
	platform_skip = ()
	tmpdir_nums = []
	test_name = None
	is_helper = False

	def __init__(self, cfg, trunner, cfgs, spawn):
		if hasattr(self, 'name'): # init will be called multiple times for classes with multiple inheritance
			return
		self.name = type(self).__name__
		self.cfg = cfg
		self.proto = cfg._proto
		self.tr = trunner
		self.cfgs = cfgs
		self.spawn = spawn
		self.have_dfl_wallet = False
		self.usr_rand_chars = (5, 30)[bool(cfg.usr_random)]
		self.usr_rand_arg = f'-r{self.usr_rand_chars}'
		self.tn_ext = ('', '.testnet')[self.proto.testnet]
		self.coin = self.proto.coin.lower()
		self.fork = 'btc' if self.coin == 'bch' and not cfg.cashaddr else self.coin
		self.altcoin_pfx = '' if self.fork == 'btc' else f'-{self.proto.coin}'
		self.testnet_opt = ['--testnet=1'] if cfg.testnet else []
		if len(self.tmpdir_nums) == 1:
			self.tmpdir_num = self.tmpdir_nums[0]
		if self.tr:
			self.spawn_env = dict(self.tr.spawn_env)
			self.spawn_env['MMGEN_TEST_SUITE_ENABLE_COLOR'] = '1' if self.color else ''
		else:
			self.spawn_env = {} # placeholder

	def get_altcoin_pfx(self, coin, cashaddr=True):
		coin = coin.lower()
		fork = 'btc' if coin == 'bch' and not cashaddr else coin
		return '' if fork == 'btc' else f'-{coin.upper()}'

	@property
	def tmpdir(self):
		return os.path.join('test', 'tmp', '{}{}'.format(self.tmpdir_num, '-α' if self.cfg.debug_utf8 else ''))

	def get_file_with_ext(self, ext, **kwargs):
		return get_file_with_ext(self.tmpdir, ext, **kwargs)

	def read_from_tmpfile(self, fn, binary=False, subdir=None):
		tdir = os.path.join(self.tmpdir, subdir) if subdir else self.tmpdir
		return read_from_file(os.path.join(tdir, fn), binary=binary)

	def write_to_tmpfile(self, fn, data, binary=False):
		return write_to_file(os.path.join(self.tmpdir, fn), data, binary=binary)

	def delete_tmpfile(self, fn):
		fn = os.path.join(self.tmpdir, fn)
		try:
			return os.unlink(fn)
		except:
			msg(f'{fn}: file does not exist or could not be deleted')

	def skip_for_platform(self, name, extra_msg=None):
		if sys.platform == name:
			msg(gray('Skipping test {!r} for {} platform{}'.format(
				self.test_name,
				name,
				f' ({extra_msg})' if extra_msg else "")))
			return True
		else:
			return False

	def skip_for_mac(self, extra_msg=None):
		return self.skip_for_platform('darwin', extra_msg)

	def skip_for_win(self, extra_msg=None):
		return self.skip_for_platform('win32', extra_msg)

	def spawn_chk(self, *args, **kwargs):
		"""
		Drop-in replacement for spawn() + t.read() for tests that spawn more than one process.
		Ensures that test script execution stops when a spawned process fails.

		"""
		t = self.spawn(*args, **kwargs)
		t.read()
		t.ok()
		t.skip_ok = True
		return t

	def noop(self):
		return 'ok'

	def _cashaddr_opt(self, val):
		return [f'--cashaddr={val}'] if self.proto.coin == 'BCH' else []

	def _kill_process_from_pid_file(self, fn, desc):
		self.spawn(msg_only=True)
		pid = int(self.read_from_tmpfile(fn))
		self.delete_tmpfile(fn)
		from signal import SIGTERM
		imsg(purple(f'Killing {desc} [PID {pid}]'))
		try:
			os.kill(pid, SIGTERM)
		except:
			imsg(yellow(f'{pid}: no such process'))
		return 'ok'



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
term: Terminal classes for the MMGen suite
"""

# TODO: reimplement as instance instead of class

import sys, os, time
from collections import namedtuple

from .util import msg, msg_r, die

if sys.platform in ('linux', 'darwin'):
	import tty, termios
	from select import select
	hold_protect_timeout = 2 if sys.platform == 'darwin' else 0.3
elif sys.platform == 'win32':
	try:
		import msvcrt
	except:
		die(2, 'Unable to set terminal mode')
	if not sys.stdin.isatty():
		msvcrt.setmode(sys.stdin.fileno(), os.O_BINARY)
else:
	die(2, f'{sys.platform!r}: unsupported platform')

_term_dimensions = namedtuple('terminal_dimensions', ['width', 'height'])

class MMGenTerm:

	@classmethod
	def register_cleanup(cls):
		pass

	@classmethod
	def init(cls, *, noecho=False):
		pass

	@classmethod
	def set(cls, *args, **kwargs):
		pass

	@classmethod
	def reset(cls):
		pass

	@classmethod
	def kb_hold_protect(cls):
		return None

class MMGenTermLinux(MMGenTerm):

	@classmethod
	def register_cleanup(cls):
		if not hasattr(cls, 'cleanup_registered'):
			import atexit
			atexit.register(
				lambda: termios.tcsetattr(
					cls.stdin_fd,
					termios.TCSADRAIN,
					cls.orig_term))
			cls.cleanup_registered = True

	@classmethod
	def reset(cls):
		termios.tcsetattr(cls.stdin_fd, termios.TCSANOW, cls.orig_term)
		cls.cur_term = cls.orig_term

	@classmethod
	def set(cls, setting):
		d = {
			'echo':   lambda t: t[:3] + [t[3] |  (termios.ECHO | termios.ECHONL)] + t[4:], # echo input chars
			'noecho': lambda t: t[:3] + [t[3] & ~(termios.ECHO | termios.ECHONL)] + t[4:], # don’t echo input chars
		}
		termios.tcsetattr(cls.stdin_fd, termios.TCSANOW, d[setting](cls.cur_term))
		cls.cur_term = termios.tcgetattr(cls.stdin_fd)

	@classmethod
	def init(cls, *, noecho=False):
		cls.stdin_fd = sys.stdin.fileno()
		cls.cur_term = termios.tcgetattr(cls.stdin_fd)
		if not hasattr(cls, 'orig_term'):
			cls.orig_term = cls.cur_term
		if noecho:
			cls.set('noecho')

	@classmethod
	def get_terminal_size(cls):
		try:
			ret = os.get_terminal_size()
		except:
			try:
				ret = (
					int(os.environ['COLUMNS']),
					int(os.environ['LINES']))
			except:
				ret = (80, 25)
		return _term_dimensions(*ret)

	@classmethod
	def kb_hold_protect(cls):
		if cls.cfg.hold_protect_disable:
			return
		tty.setcbreak(cls.stdin_fd)
		while True:
			key = select([sys.stdin], [], [], hold_protect_timeout)[0]
			if key:
				sys.stdin.read(1)
			else:
				termios.tcsetattr(cls.stdin_fd, termios.TCSADRAIN, cls.cur_term)
				break

	@classmethod
	def get_char(cls, prompt='', *, immed_chars='', prehold_protect=True, num_bytes=5):
		"""
		Use os.read(), not file.read(), to get a variable number of bytes without blocking.
		Request 5 bytes to cover escape sequences generated by F1, F2, .. Fn keys (5 bytes)
		as well as UTF8 chars (4 bytes max).
		"""
		timeout = 0.3
		tty.setcbreak(cls.stdin_fd)
		msg_r(prompt)
		if cls.cfg.hold_protect_disable:
			prehold_protect = False
		while True:
			# Protect against held-down key before read()
			key = select([sys.stdin], [], [], timeout)[0]
			s = os.read(cls.stdin_fd, num_bytes).decode()
			if prehold_protect and key:
				continue
			if s in immed_chars:
				break
			# Protect against long keypress
			key = select([sys.stdin], [], [], timeout)[0]
			if not key:
				break
		termios.tcsetattr(cls.stdin_fd, termios.TCSADRAIN, cls.cur_term)
		return s

	@classmethod
	def get_char_raw(cls, prompt='', num_bytes=5, **kwargs):
		tty.setcbreak(cls.stdin_fd)
		msg_r(prompt)
		s = os.read(cls.stdin_fd, num_bytes).decode()
		termios.tcsetattr(cls.stdin_fd, termios.TCSADRAIN, cls.cur_term)
		return s

class MMGenTermLinuxStub(MMGenTermLinux):

	@classmethod
	def register_cleanup(cls):
		pass

	@classmethod
	def init(cls, *, noecho=False):
		cls.stdin_fd = sys.stdin.fileno()

	@classmethod
	def set(cls, *args, **kwargs):
		pass

	@classmethod
	def reset(cls):
		pass

	@classmethod
	def get_char(cls, prompt='', *, immed_chars='', prehold_protect=None, num_bytes=5):
		msg_r(prompt)
		return os.read(0, num_bytes).decode()

	get_char_raw = get_char

	@classmethod
	def kb_hold_protect(cls):
		pass

class MMGenTermMSWin(MMGenTerm):

	@classmethod
	def get_terminal_size(cls):
		import struct
		x, y = 0, 0
		try:
			from ctypes import windll, create_string_buffer
			# handles - stdin: -10, stdout: -11, stderr: -12
			csbi = create_string_buffer(22)
			h = windll.kernel32.GetStdHandle(-12)
			res = windll.kernel32.GetConsoleScreenBufferInfo(h, csbi)
			assert res, 'failed to get console screen buffer info'
			left, top, right, bottom = struct.unpack('hhhhHhhhhhh', csbi.raw)[5:9]
			x = right - left + 1
			y = bottom - top + 1
		except:
			pass

		if x and y:
			return _term_dimensions(x, y)
		else:
			from .color import yellow
			msg(yellow('Warning: could not get terminal size. Using fallback dimensions.'))
			return _term_dimensions(80, 25)

	@classmethod
	def kb_hold_protect(cls):
		timeout = 0.5
		while True:
			hit_time = time.time()
			while True:
				if msvcrt.kbhit():
					msvcrt.getch()
					break
				if time.time() - hit_time > timeout:
					return

	@classmethod
	def get_char(cls, prompt='', *, immed_chars='', prehold_protect=True, num_bytes=None):
		"""
		always return a single character, ignore num_bytes
		first character of 2-character sequence returned by F1-F12 keys is discarded
		prehold_protect is ignored
		"""
		msg_r(prompt)
		timeout = 0.5
		while True:
			if msvcrt.kbhit():
				ch = chr(msvcrt.getch()[0])
				if ch == '\x03':
					raise KeyboardInterrupt
				if ch in immed_chars:
					return ch
				hit_time = time.time()
				while True:
					if msvcrt.kbhit():
						break
					if time.time() - hit_time > timeout:
						return ch

	@classmethod
	def get_char_raw(cls, prompt='', num_bytes=None, **kwargs):
		"""
		return single ASCII char or 2-char escape sequence, ignoring num_bytes
		"""
		msg_r(prompt)
		ret = msvcrt.getch()
		if ret in (b'\x00', b'\xe0'): # first byte of 2-byte escape sequence
			return chr(ret[0]) + chr(msvcrt.getch()[0])
		if ret == b'\x03':
			raise KeyboardInterrupt
		return chr(ret[0])

class MMGenTermMSWinStub(MMGenTermMSWin):

	@classmethod
	def get_char(cls, prompt='', *, immed_chars='', prehold_protect=None, num_bytes=None):
		"""
		Use stdin to allow UTF-8 and emulate the one-character behavior of MMGenTermMSWin
		"""
		msg_r(prompt)
		while True:
			try:
				return sys.stdin.read(1)
			except:
				msg('[read error, trying again]')
				time.sleep(0.5)

	get_char_raw = get_char

def get_term():
	return {
		'linux': (MMGenTermLinux if sys.stdin.isatty() else MMGenTermLinuxStub),
		'darwin': (MMGenTermLinux if sys.stdin.isatty() else MMGenTermLinuxStub),
		'win32': (MMGenTermMSWin if sys.stdin.isatty() else MMGenTermMSWinStub),
	}[sys.platform]

def init_term(cfg, *, noecho=False):

	term = get_term()

	term.init(noecho=noecho)

	for var in ('get_char', 'get_char_raw', 'kb_hold_protect', 'get_terminal_size'):
		globals()[var] = getattr(term, var)

	term.cfg = cfg # setting the _class_ attribute

def reset_term():
	get_term().reset()



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
proto.eth.tw.ctl: Ethereum tracking wallet control class
"""

from ....util import msg, ymsg, die, cached_property
from ....tw.store import TwCtlWithStore
from ....tw.ctl import write_mode
from ....addr import is_coin_addr

from ..contract import Token

class EthereumTwCtl(TwCtlWithStore):

	data_key = 'accounts'

	def init_empty(self):
		self.data = {
			'coin': self.proto.coin,
			'network': self.proto.network.upper(),
			'accounts': {},
			'tokens': {},
		}

	def upgrade_wallet_maybe(self):

		upgraded = False

		if not 'accounts' in self.data or not 'coin' in self.data:
			ymsg(f'Upgrading {self.desc} (v1->v2: accounts field added)')
			if not 'accounts' in self.data:
				self.data = {}
				import json
				self.data['accounts'] = json.loads(self.orig_data)
			if not 'coin' in self.data:
				self.data['coin'] = self.proto.coin
			upgraded = True

		def have_token_params_fields():
			for k in self.data['tokens']:
				if 'params' in self.data['tokens'][k]:
					return True

		def add_token_params_fields():
			for k in self.data['tokens']:
				self.data['tokens'][k]['params'] = {}

		if not 'tokens' in self.data:
			self.data['tokens'] = {}
			upgraded = True

		if self.data['tokens'] and not have_token_params_fields():
			ymsg(f'Upgrading {self.desc} (v2->v3: token params fields added)')
			add_token_params_fields()
			upgraded = True

		if not 'network' in self.data:
			ymsg(f'Upgrading {self.desc} (v3->v4: network field added)')
			self.data['network'] = self.proto.network.upper()
			upgraded = True

		if upgraded:
			self.force_write()
			msg(f'{self.desc} upgraded successfully!')

	async def rpc_get_balance(self, addr, block='latest'):
		return self.proto.coin_amt(
			int(await self.rpc.call('eth_getBalance', '0x' + addr, block), 16),
			from_unit = 'wei')

	async def addr2sym(self, req_addr):
		for addr in self.data['tokens']:
			if addr == req_addr:
				return self.data['tokens'][addr]['params']['symbol']

	async def sym2addr(self, sym):
		for addr in self.data['tokens']:
			if self.data['tokens'][addr]['params']['symbol'].upper() == sym.upper():
				return addr

	# Since it’s nearly impossible to empty an Ethereum account, consider set of used addresses
	# to be all accounts with balances.
	# Token addresses might have a balance but no corresponding ETH balance, so check them too.
	@cached_property
	def used_addrs(self):
		from decimal import Decimal
		return (
			{k for k, v in self.data['accounts'].items() if Decimal(v.get('balance', 0))} |
			{k for t in self.data['tokens'].values() for k, v in t.items()
				if Decimal(v.get('balance', 0))})

class EthereumTokenTwCtl(EthereumTwCtl):

	desc = 'Ethereum token tracking wallet'
	decimals = None
	symbol = None
	cur_eth_balances = {}

	async def __init__(self, cfg, proto, *, mode='r', token_addr=None, no_rpc=False):

		await super().__init__(cfg, proto, mode=mode, no_rpc=no_rpc)

		for v in self.data['tokens'].values():
			self.conv_types(v)

		if self.importing and token_addr:
			if not is_coin_addr(proto, token_addr):
				die('InvalidContractAddress', f'{token_addr!r}: invalid token address')
		else:
			assert token_addr is None, 'EthereumTokenTwCtl_chk1'
			token_addr = await self.sym2addr(proto.tokensym) # returns None on failure
			if not is_coin_addr(proto, token_addr):
				die('UnrecognizedTokenSymbol', f'Specified token {proto.tokensym!r} could not be resolved!')

		from ....addr import ContractAddr
		self.token = ContractAddr(proto, token_addr)

		if self.token not in self.data['tokens']:
			if self.importing:
				await self.import_token(self.token)
			else:
				die('TokenNotInWallet', f'Specified token {self.token!r} not in wallet!')

		self.decimals = self.get_param('decimals')
		self.symbol   = self.get_param('symbol')
		if mode == 'i' and not proto.tokensym:
			proto.tokensym = self.symbol

	@property
	def data_root(self):
		return self.data['tokens'][self.token]

	@property
	def data_root_desc(self):
		return 'token ' + self.get_param('symbol')

	async def rpc_get_balance(self, addr, block='latest'):
		return await Token(
			self.cfg,
			self.proto,
			self.token,
			decimals = self.decimals,
			rpc = self.rpc).get_balance(addr, block=block)

	async def get_eth_balance(self, addr, *, force_rpc=False, block='latest'):
		cache = self.cur_eth_balances
		r = self.data['accounts']
		ret = None if force_rpc else self.get_cached_balance(addr, cache, r)
		if ret is None:
			ret = await super().rpc_get_balance(addr, block=block)
			self.cache_balance(addr, ret, session_cache=cache, data_root=r)
		return ret

	def get_param(self, param):
		return self.data['tokens'][self.token]['params'][param]

	@write_mode
	async def import_token(self, tokenaddr):
		"""
		Token 'symbol' and 'decimals' values are resolved from the network by the system just
		once, upon token import.  Thereafter, token address, symbol and decimals are resolved
		either from the tracking wallet (online operations) or transaction file (when signing).
		"""
		t = Token(self.cfg, self.proto, tokenaddr, rpc=self.rpc)
		self.data['tokens'][tokenaddr] = {
			'params': {
				'symbol': await t.get_symbol(),
				'decimals': await t.get_decimals()
			}
		}



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
tx.new_swap: new swap transaction class
"""

from collections import namedtuple

from .new import New
from ..amt import UniAmt

def get_swap_proto_mod(swap_proto_name):
	import importlib
	return importlib.import_module(f'mmgen.swap.proto.{swap_proto_name}')

def init_swap_proto(cfg, asset):
	from ..protocol import init_proto
	return init_proto(
		cfg,
		asset.coin,
		network = cfg._proto.network,
		tokensym = asset.tokensym,
		need_amt = True)

def get_send_proto(cfg):
	try:
		arg = cfg._args.pop(0)
	except:
		cfg._usage()
	return init_swap_proto(cfg, get_swap_proto_mod(cfg.swap_proto).SwapAsset(arg, 'send'))

class NewSwap(New):
	desc = 'swap transaction'
	swap_quote_refresh_timeout = 30

	def __init__(self, *args, **kwargs):
		self.is_swap = True
		self.swap_proto = kwargs['cfg'].swap_proto
		New.__init__(self, *args, **kwargs)

	def check_addr_is_wallet_addr(self, output, *, message):
		if not output.mmid:
			self._non_wallet_addr_confirm(message)

	async def get_swap_output(self, proto, arg, addrfiles, desc):
		ret = namedtuple('swap_output', ['coin', 'network', 'addr', 'mmid'])
		if arg:
			from ..addrdata import TwAddrData
			pa = self.parse_cmdline_arg(
				proto,
				arg,
				self.get_addrdata_from_files(proto, addrfiles),
				await TwAddrData(self.cfg, proto))
			if pa.addr:
				await self.warn_addr_used(proto, pa, desc)
				return ret(proto.coin, proto.network, pa.addr, pa.mmid)

		full_desc = f'{desc} on the {proto.coin} {proto.network} network'
		res = await self.get_autochg_addr(proto, arg, exclude=[], desc=full_desc, all_addrtypes=not arg)
		self.confirm_autoselected_addr(res.twmmid, full_desc)
		return ret(proto.coin, proto.network, res.addr, res.twmmid)

	async def get_chg_output(self, arg, addrfiles):
		chg_output = await self.get_swap_output(self.proto, arg, addrfiles, 'change address')
		self.check_addr_is_wallet_addr(
			chg_output,
			message = 'Change address is not an MMGen wallet address!')
		return chg_output

	async def process_swap_cmdline_args(self, cmd_args, addrfiles):

		class CmdlineArgs: # listed in command-line order
			# send_coin       # required: uppercase coin symbol
			send_amt  = None  # optional: Omit to skip change addr and send value of all inputs minus fees
							  #           to vault
			# chg_spec = None # optional: change address spec, e.g. ‘B’ ‘DEADBEEF:B’ ‘DEADBEEF:B:1’ or
							  #           coin address.  Omit for autoselected change address.  Use of
							  #           non-wallet change address will emit warning and prompt user
							  #           for confirmation
			# recv_coin       # required: uppercase coin symbol
			recv_spec = None  # optional: destination address spec. Same rules as for chg_spec

		def get_arg():
			try:
				return args_in.pop(0)
			except:
				self.cfg._usage()

		async def parse():

			# arg 1: send_coin - already popped and parsed by get_send_proto()

			from ..amt import is_coin_amt
			arg = get_arg()

			# arg 2: amt
			if is_coin_amt(self.proto, arg):
				UniAmt(arg) # throw exception on decimal overflow
				args.send_amt = self.proto.coin_amt(arg)
				arg = get_arg()

			# arg 3: chg_spec (change address spec)
			if args.send_amt and not (self.proto.is_vm or arg in sa.recv): # is change arg
				nonlocal chg_output
				chg_output = await self.get_chg_output(arg, addrfiles)
				arg = get_arg()

			# arg 4: recv_coin
			self.swap_recv_asset_spec = arg # this goes into the transaction file
			self.recv_proto = init_swap_proto(self.cfg, self.recv_asset)

			# arg 5: recv_spec (receive address spec)
			if args_in:
				args.recv_spec = get_arg()

			if args_in: # done parsing, all args consumed
				self.cfg._usage()

		sp = self.swap_proto_mod
		sa = sp.SwapAsset('BTC', 'send')
		args_in = list(cmd_args)
		args = CmdlineArgs()
		chg_output = None

		await parse()

		for a in (self.send_asset, self.recv_asset):
			if a.name not in sa.tested:
				from ..util import msg, ymsg
				from ..term import get_char
				ymsg(f'Warning: {a.direction} asset {a.name} is untested by the MMGen Project')
				get_char('Press any key to continue: ')
				msg('')

		if args.send_amt and not (chg_output or self.proto.is_vm):
			chg_output = await self.get_chg_output(None, addrfiles)

		recv_output = await self.get_swap_output(
			self.recv_proto,
			args.recv_spec,
			addrfiles,
			'destination address')

		self.check_addr_is_wallet_addr(
			recv_output,
			message = (
				'Swap destination address is not an MMGen wallet address!\n'
				'To sign this transaction, autosign or txsign must be invoked'
				' with --allow-non-wallet-swap'))

		sc = self.swap_cfg = self.swap_proto_mod.SwapCfg(self.cfg)

		memo = sp.Memo(
			self.swap_cfg,
			self.recv_proto,
			self.recv_asset,
			recv_output.addr,
			# sc.trade_limit could be a float:
			trade_limit = sc.trade_limit if isinstance(sc.trade_limit, UniAmt) else None)

		# this goes into the transaction file:
		self.swap_recv_addr_mmid = recv_output.mmid

		return (
			[f'vault,{args.send_amt}', f'data:{memo}'] if args.send_amt and self.proto.is_vm else
			[f'vault,{args.send_amt}', chg_output.mmid, f'data:{memo}'] if args.send_amt else
			['vault', f'data:{memo}'])

	def update_vault_addr(self, c, *, addr='inbound_address'):
		vault_idx = self.vault_idx
		assert vault_idx == 0, f'{vault_idx}: vault index is not zero!'
		o = self.outputs[vault_idx]._asdict()
		o['addr'] = getattr(c, addr)
		self.outputs[vault_idx] = self.Output(self.proto, **o)

	async def update_vault_output(self, amt, *, deduct_est_fee=False):
		c = self.swap_proto_mod.rpc_client(self, amt)

		import time
		from ..util import msg
		from ..term import get_char

		def get_trade_limit():
			if type(self.swap_cfg.trade_limit) is UniAmt:
				return self.swap_cfg.trade_limit
			elif type(self.swap_cfg.trade_limit) is float:
				return (
					UniAmt(int(c.data['expected_amount_out']), from_unit='satoshi')
					* self.swap_cfg.trade_limit)

		while True:
			self.cfg._util.qmsg(f'Retrieving data from {c.rpc.host}...')
			c.get_quote(self.swap_cfg)
			self.cfg._util.qmsg('OK')
			self.swap_quote_refresh_time = time.time()
			await self.set_gas(to_addr=c.router if self.is_token else None)
			trade_limit = get_trade_limit()
			msg(await c.format_quote(trade_limit, deduct_est_fee=deduct_est_fee))
			ch = get_char('Press ‘r’ to refresh quote, any other key to continue: ')
			msg('')
			if ch not in 'Rr':
				break

		self.swap_quote_expiry = c.data['expiry']
		self.update_vault_addr(c)
		self.update_data_output(trade_limit)
		self.quote_data = c
		return c.rel_fee_hint



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
tw.bal: Tracking wallet getbalance class for the MMGen suite
"""

from ..base_obj import AsyncInit
from ..objmethods import MMGenObject
from ..obj import NonNegativeInt

class TwGetBalance(MMGenObject, metaclass=AsyncInit):

	def __new__(cls, cfg, proto, *, minconf, quiet):
		return MMGenObject.__new__(proto.base_proto_subclass(cls, 'tw.bal'))

	async def __init__(self, cfg, proto, *, minconf, quiet):

		class BalanceInfo(dict):
			def __init__(self):
				amt0 = proto.coin_amt('0')
				data = {
					'unconfirmed': amt0,
					'lt_minconf': amt0,
					'ge_minconf': amt0,
					'spendable': amt0,
				}
				dict.__init__(self, **data)

		self.minconf = NonNegativeInt(minconf)
		self.balance_info = BalanceInfo
		self.quiet = quiet
		self.proto = proto
		self.data = {k:self.balance_info() for k in self.start_labels}

		if minconf < 2 and 'lt_minconf' in self.conf_cols:
			del self.conf_cols['lt_minconf']

		await self.create_data()

	def format(self, color):

		def gen_output():

			if self.quiet:
				yield str(self.data['TOTAL']['ge_minconf'] if self.data else 0)
			else:

				def get_col_iwidth(colname):
					return len(str(int(max(v[colname] for v in self.data.values())))) + iwidth_adj

				def make_col(label, col):
					return self.data[label][col].fmt(iwidths[col], color=color)

				if color:
					from ..color import green, yellow
				else:
					from ..color import nocolor
					green = yellow = nocolor

				add_w = self.proto.coin_amt.max_prec + 1 # 1 = len('.')
				iwidth_adj = 1 # so that min iwidth (1) + add_w + iwidth_adj >= len('Unconfirmed')
				col1_w = max(len(l) for l in self.start_labels) + 1 # 1 = len(':')

				iwidths = {colname: get_col_iwidth(colname) for colname in self.conf_cols}

				net_desc = self.proto.coin + ' ' + self.proto.network.upper()
				if net_desc != 'BTC MAINNET':
					yield f'Network: {green(net_desc)}'

				yield '{lbl:{w}} {cols}'.format(
					lbl = 'Wallet',
					w = col1_w + iwidth_adj,
					cols = ' '.join(v.format(minconf=self.minconf).ljust(iwidths[k]+add_w)
						for k, v in self.conf_cols.items())).rstrip()

				from ..addr import MMGenID
				for label in sorted(self.data.keys()):
					yield '{lbl} {cols}'.format(
						lbl = yellow((label + ' ' + self.proto.coin).ljust(col1_w)) if label == 'TOTAL'
							else MMGenID.hlc((label+':').ljust(col1_w), color=color),
						cols = ' '.join(make_col(label, col) for col in self.conf_cols)
					).rstrip()

		return '\n'.join(gen_output())



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>

"""
test.objtest_d.eth_mainnet: ETH mainnet test vectors for MMGen data objects
"""

from decimal import Decimal

from mmgen.obj import ETHNonce
from mmgen.amt import ETHAmt
from mmgen.protocol import init_proto
from mmgen.addr import CoinAddr

from ..include.common import cfg

proto = init_proto(cfg, 'eth', need_amt=True)

tests = {
	'CoinAddr': {
		'arg1': 'addr',
		'good': (
			{'addr': 'beadcafe' * 5, 'proto': proto},
		),
		'bad': (
			{'addr': 'aaaaxxxx' * 5, 'proto': proto},
			{'addr': 'beadcafe' * 2, 'proto': proto},
			{'addr': 'beadcafe' * 6, 'proto': proto},
		),
	},
	'ETHAmt': {
		'bad': ('-3.2', '0.1234567891234567891', 123, '123L',
					{'num': '1', 'from_decimal': True},
					{'num': 1, 'from_decimal': True},
				),
		'good': (('123.123456789123456789', Decimal('123.123456789123456789')),
				{   'num': Decimal('1.12345678912345678892345'), # rounding
					'from_decimal': True,
					'ret': Decimal('1.123456789123456789')},
				{'num': Decimal('1.234'), 'from_decimal': True, 'ret': Decimal('1.234')},
				{'num': Decimal('0.0'), 'from_decimal': True, 'ret': Decimal('0')},
				{'num': 1234, 'from_unit': 'wei', 'ret': Decimal('0.000000000000001234')},
				{'num': 1234, 'from_unit': 'Mwei', 'ret': Decimal('0.000000001234')},
		)
	},
	'ETHNonce': {
		'bad': ('z', 'я', -1, '-1', 0.0, '0.0'),
		'good': (
			('0', 0), ('1', 1), ('100', 100), 1, 100,
			{'n': '0x0', 'base': 16, 'ret': 0},
			{'n': '0x1', 'base': 16, 'ret': 1},
			{'n': '0xf', 'base': 16, 'ret': 15},
			{'n': '0xff', 'base': 16, 'ret': 255},
		)
	},
}



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.btc.daemon: Bitcoin base protocol daemon classes
"""

import os

from ...cfg import gc
from ...util import list_gen
from ...daemon import CoinDaemon, _nw, _dd

class bitcoin_core_daemon(CoinDaemon):
	daemon_data = _dd('Bitcoin Core', 290100, '29.1.0')
	exec_fn = 'bitcoind'
	cli_fn = 'bitcoin-cli'
	testnet_dir = 'testnet3'
	cfg_file_hdr = '# Bitcoin Core config file\n'
	rpc_ports = _nw(8332, 18332, 18443)
	cfg_file = 'bitcoin.conf'
	nonstd_datadir = False
	datadirs = {
		'linux': [gc.home_dir, '.bitcoin'],
		'darwin': [gc.home_dir, 'Library', 'Application Support', 'Bitcoin'],
		'win32': [os.getenv('APPDATA'), 'Bitcoin']
	}
	avail_opts = ('no_daemonize', 'online', 'bdb_wallet')

	def init_datadir(self):
		if self.network == 'regtest' and not self.test_suite:
			return os.path.join(self.cfg.data_dir_root, 'regtest', self.cfg.coin.lower())
		else:
			return super().init_datadir()

	@property
	def network_datadir(self):
		"location of the network's blockchain data and authentication cookie"
		return os.path.join (
			self.datadir, {
				'mainnet': '',
				'testnet': self.testnet_dir,
				'regtest': 'regtest',
			}[self.network])

	@property
	def auth_cookie_fn(self):
		return os.path.join(self.network_datadir, '.cookie')

	def init_subclass(self):

		if self.network == 'regtest':
			# fall back on hard-coded credentials:
			from .regtest import MMGenRegtest
			self.rpc_user = MMGenRegtest.rpc_user
			self.rpc_password = MMGenRegtest.rpc_password

		self.shared_args = list_gen(
			[f'--datadir={self.datadir}',         self.nonstd_datadir or self.non_dfl_datadir],
			[f'--rpcport={self.rpc_port}'],
			[f'--rpcuser={self.rpc_user}',         self.network == 'regtest'],
			[f'--rpcpassword={self.rpc_password}', self.network == 'regtest'],
			['--testnet',                          self.network == 'testnet'],
			['--regtest',                          self.network == 'regtest'],
		)

		self.coind_args = list_gen(
			['--listen=0'],
			['--keypool=1'],
			['--rpcallowip=127.0.0.1'],
			[f'--rpcbind=127.0.0.1:{self.rpc_port}'],
			['--pid='+self.pidfile,    self.use_pidfile],
			['--daemon',               self.platform in ('linux', 'darwin') and not self.opt.no_daemonize],
			['--fallbackfee=0.0002',   self.coin == 'BTC' and self.network == 'regtest'],
			['--deprecatedrpc=create_bdb', self.coin == 'BTC' and self.opt.bdb_wallet],
			['--mempoolreplacement=1', self.coin == 'LTC'],
			['--txindex=1',            self.coin == 'LTC' or self.network == 'regtest'],
			['--addresstype=bech32',   self.coin == 'LTC' and self.network == 'regtest'],
		)

		self.lockfile = os.path.join(self.network_datadir, '.cookie')

	@property
	def state(self):
		cp = self.cli('getblockcount', silent=True)
		err = cp.stderr.decode()
		if ("error: couldn't connect" in err
			or "error: Could not connect" in err
			or "does not exist" in err):
			# regtest has no cookie file, so test will always fail
			ret = 'busy' if (self.lockfile and os.path.exists(self.lockfile)) else 'stopped'
		elif cp.returncode == 0:
			ret = 'ready'
		else:
			ret = 'busy'
		if self.debug:
			print(f'State: {ret!r}')
		return ret

	@property
	def stop_cmd(self):
		return self.cli_cmd('stop')

	def set_comment_args(self, rpc, coinaddr, lbl):
		if 'label_api' in rpc.caps:
			return ('setlabel', coinaddr, lbl)
		else:
			# NOTE: this works because importaddress() removes the old account before
			# associating the new account with the address.
			# RPC args: addr, label, rescan[=true], p2sh[=none]
			return ('importaddress', coinaddr, lbl, False)

	def estimatefee_args(self, rpc):
		return (self.cfg.fee_estimate_confs,)

	def sigfail_errmsg(self, e):
		return e.args[0]

class bitcoin_cash_node_daemon(bitcoin_core_daemon):
	daemon_data = _dd('Bitcoin Cash Node', 28000100, '28.0.1')
	exec_fn = 'bitcoind-bchn'
	cli_fn = 'bitcoin-cli-bchn'
	rpc_ports = _nw(8432, 18432, 18543) # use non-standard ports (core+100)
	cfg_file_hdr = '# Bitcoin Cash Node config file\n'
	nonstd_datadir = True
	datadirs = {
		'linux': [gc.home_dir, '.bitcoin-bchn'],
		'darwin': [gc.home_dir, 'Library', 'Application Support', 'Bitcoin-Cash-Node'],
		'win32': [os.getenv('APPDATA'), 'Bitcoin-Cash-Node']
	}

	def set_comment_args(self, rpc, coinaddr, lbl):
		# bitcoin-{abc, bchn} 'setlabel' RPC is broken, so use old 'importaddress' method to set label
		# Broken behavior: new label is set OK, but old label gets attached to another address
		return ('importaddress', coinaddr, lbl, False)

	def estimatefee_args(self, rpc):
		return () if rpc.daemon_version >= 190100 else (self.cfg.fee_estimate_confs,)

	def sigfail_errmsg(self, e):
		return (
			'This is not the BCH chain.\nRe-run the script without the --coin=bch option.'
				if 'Invalid sighash param' in e.args[0] else
			e.args[0])

class litecoin_core_daemon(bitcoin_core_daemon):
	# v0.21.2rc5 crashes when mining more than 431 blocks in regtest mode:
	#   CreateNewBlock: TestBlockValidity failed: bad-txns-vin-empty, Transaction check failed
	daemon_data = _dd('Litecoin Core', 210400, '0.21.4')
	exec_fn = 'litecoind'
	cli_fn = 'litecoin-cli'
	testnet_dir = 'testnet4'
	rpc_ports = _nw(9332, 19332, 19443)
	cfg_file = 'litecoin.conf'
	cfg_file_hdr = '# Litecoin Core config file\n'
	datadirs = {
		'linux': [gc.home_dir, '.litecoin'],
		'darwin': [gc.home_dir, 'Library', 'Application Support', 'Litecoin'],
		'win32': [os.getenv('APPDATA'), 'Litecoin']
	}



import os as overlay_fake_os
from .unspent_orig import *

if overlay_fake_os.getenv('MMGEN_BOGUS_UNSPENT_DATA'):

	class overlay_fake_BitcoinTwUnspentOutputs(BitcoinTwUnspentOutputs):

		async def get_rpc_data(self):
			import json
			from ....fileutil import get_data_from_file
			return json.loads(get_data_from_file(
				self.cfg,
				overlay_fake_os.getenv('MMGEN_BOGUS_UNSPENT_DATA')
			))

	BitcoinTwUnspentOutputs.get_rpc_data = overlay_fake_BitcoinTwUnspentOutputs.get_rpc_data



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
util: Frequently-used variables, classes and utility functions for the MMGen suite
"""

import sys, os, time, re

from .color import red, yellow, green, blue, purple
from .cfg import gv

ascii_lowercase = 'abcdefghijklmnopqrstuvwxyz'

digits = '0123456789'
hexdigits = '0123456789abcdefABCDEF'
hexdigits_uc = '0123456789ABCDEF'
hexdigits_lc = '0123456789abcdef'

def noop(*args, **kwargs):
	pass

class Util:

	def __init__(self, cfg):

		self.cfg = cfg

		if cfg.quiet:
			self.qmsg = self.qmsg_r = noop
		else:
			self.qmsg = msg
			self.qmsg_r = msg_r

		if cfg.verbose:
			self.vmsg = msg
			self.vmsg_r = msg_r
			self.Vmsg = Msg
			self.Vmsg_r = Msg_r
		else:
			self.vmsg = self.vmsg_r = self.Vmsg = self.Vmsg_r = noop

		self.dmsg = msg if cfg.debug else noop

		if cfg.pager:
			from .ui import do_pager
			self.stdout_or_pager = do_pager
		else:
			self.stdout_or_pager = Msg_r

	def compare_chksums(
			self,
			chk1,
			desc1,
			chk2,
			desc2,
			*,
			hdr         = '',
			die_on_fail = False,
			verbose     = False):

		if not chk1 == chk2:
			fs = "{} ERROR: {} checksum ({}) doesn't match {} checksum ({})"
			m = fs.format((hdr+':\n   ' if hdr else 'CHECKSUM'), desc2, chk2, desc1, chk1)
			if die_on_fail:
				die(3, m)
			else:
				if verbose or self.cfg.verbose:
					msg(m)
				return False

		if self.cfg.verbose:
			msg(f'{capfirst(desc1)} checksum OK ({chk1})')

		return True

	def compare_or_die(self, val1, desc1, val2, desc2, *, e='Error'):
		if val1 != val2:
			die(3, f"{e}: {desc2} ({val2}) doesn't match {desc1} ({val1})")
		if self.cfg.debug:
			msg(f'{capfirst(desc2)} OK ({val2})')
		return True

if sys.platform == 'win32':
	def msg_r(s):
		try:
			gv.stderr.write(s)
			gv.stderr.flush()
		except:
			os.write(2, s.encode())

	def msg(s):
		msg_r(s + '\n')

	def Msg_r(s):
		try:
			gv.stdout.write(s)
			gv.stdout.flush()
		except:
			os.write(1, s.encode())

	def Msg(s):
		Msg_r(s + '\n')
else:
	def msg(s):
		gv.stderr.write(s + '\n')

	def msg_r(s):
		gv.stderr.write(s)
		gv.stderr.flush()

	def Msg(s):
		gv.stdout.write(s + '\n')

	def Msg_r(s):
		gv.stdout.write(s)
		gv.stdout.flush()

def rmsg(s):
	msg(red(s))

def ymsg(s):
	msg(yellow(s))

def gmsg(s):
	msg(green(s))

def gmsg_r(s):
	msg_r(green(s))

def bmsg(s):
	msg(blue(s))

def pumsg(s):
	msg(purple(s))

def mmsg(*args):
	for d in args:
		Msg(repr(d))

def mdie(*args):
	mmsg(*args)
	sys.exit(0)

def die(ev, s='', *, stdout=False):
	if isinstance(ev, int):
		from .exception import MMGenSystemExit, MMGenError
		if ev <= 2:
			raise MMGenSystemExit(ev, s, stdout)
		else:
			raise MMGenError(ev, s, stdout)
	elif isinstance(ev, str):
		from . import exception
		raise getattr(exception, ev)(s)
	else:
		raise ValueError(f'{ev}: exit value must be string or int instance')

def Die(ev=0, s=''):
	die(ev=ev, s=s, stdout=True)

def pp_fmt(d):
	import pprint
	return pprint.PrettyPrinter().pformat(d)

def pp_msg(d):
	msg(pp_fmt(d))

def indent(s, *, indent='    ', append='\n'):
	"indent multiple lines of text with specified string"
	return indent + ('\n'+indent).join(s.strip().splitlines()) + append

def fmt(s, *, indent='', strip_char=None, append='\n'):
	"de-indent multiple lines of text, or indent with specified string"
	return indent + ('\n'+indent).join([l.lstrip(strip_char) for l in s.strip().splitlines()]) + append

def fmt_list(iterable, *, fmt='dfl', indent='', conv=None):
	"pretty-format a list"
	_conv, sep, lq, rq = {
		'dfl':       (str,  ", ", "'",  "'"),
		'utf8':      (str,  ", ", "“",  "”"),
		'bare':      (repr, " ",  "",   ""),
		'barest':    (str,  " ",  "",   ""),
		'fancy':     (str,  " ",  "‘",  "’"),
		'no_quotes': (str,  ", ", "",   ""),
		'compact':   (str,  ",",  "",   ""),
		'no_spc':    (str,  ",",  "'",  "'"),
		'min':       (str,  ",",  "",   ""),
		'repr':      (repr, ", ", "",   ""),
		'csv':       (repr, ",",  "",   ""),
		'col':       (str,  "\n", "",   ""),
	}[fmt]
	conv = conv or _conv
	return indent + (sep+indent).join(lq+conv(e)+rq for e in iterable)

def fmt_dict(mapping, *, fmt='dfl', kconv=None, vconv=None):
	"pretty-format a dict"
	kc, vc, sep, fs = {
		'dfl':           (str, str,  ", ",  "'{}' ({})"),
		'dfl_compact':   (str, str,  " ",   "{} ({})"),
		'square':        (str, str,  ", ",  "'{}' [{}]"),
		'square_compact':(str, str,  " ",   "{} [{}]"),
		'equal':         (str, str,  ", ",  "'{}'={}"),
		'equal_spaced':  (str, str,  ", ",  "'{}' = {}"),
		'equal_compact': (str, str,  " ",   "{}={}"),
		'kwargs':        (str, repr, ", ",  "{}={}"),
		'colon':         (str, repr, ", ",  "{}:{}"),
		'colon_compact': (str, str,  " ",   "{}:{}"),
	}[fmt]
	kconv = kconv or kc
	vconv = vconv or vc
	return sep.join(fs.format(kconv(k), vconv(v)) for k, v in mapping.items())

def list_gen(*data):
	"""
	Generate a list from an arg tuple of sublists
	- The last element of each sublist is a condition.  If it evaluates to true, the preceding
	  elements of the sublist are included in the result.  Otherwise the sublist is skipped.
	- If a sublist contains only one element, the condition defaults to true.
	"""
	assert type(data) in (list, tuple), f'{type(data).__name__} not in (list, tuple)'
	def gen():
		for d in data:
			assert isinstance(d, list), f'{type(d).__name__} != list'
			if len(d) == 1:
				yield d[0]
			elif d[-1]:
				for idx in range(len(d)-1):
					yield d[idx]
	return list(gen())

def remove_dups(iterable, *, edesc='element', desc='list', quiet=False, hide=False):
	"""
	Remove duplicate occurrences of iterable elements, preserving first occurrence
	If iterable is a generator, return a list, else type(iterable)
	"""
	ret = []
	for e in iterable:
		if e in ret:
			if not quiet:
				ymsg(f'Warning: removing duplicate {edesc} {"(hidden)" if hide else e} in {desc}')
		else:
			ret.append(e)
	return ret if type(iterable).__name__ == 'generator' else type(iterable)(ret)

def contains_any(target_list, source_list):
	return any(map(target_list.count, source_list))

def suf(arg, suf_type='s', *, verb='none'):
	suf_types = {
		'none': {
			's':   ('s',   ''),
			'es':  ('es',  ''),
			'ies': ('ies', 'y'),
		},
		'is': {
			's':   ('s are',   ' is'),
			'es':  ('es are',  ' is'),
			'ies': ('ies are', 'y is'),
		},
		'has': {
			's':   ('s have',   ' has'),
			'es':  ('es have',  ' has'),
			'ies': ('ies have', 'y has'),
		},
	}
	if isinstance(arg, int):
		n = arg
	elif isinstance(arg, (list, tuple, set, dict)):
		n = len(arg)
	else:
		die(2, f'{arg}: invalid parameter for suf()')
	return suf_types[verb][suf_type][n == 1]

def get_extension(fn):
	return os.path.splitext(fn)[1][1:]

def remove_extension(fn, ext):
	a, b = os.path.splitext(fn)
	return a if b[1:] == ext else fn

def make_chksum_N(s, nchars, *, sep=False, rounds=2, upper=True):
	if isinstance(s, str):
		s = s.encode()
	from hashlib import sha256
	for i in range(rounds):
		s = sha256(s).digest()
	ret = s.hex()[:nchars]
	if sep:
		assert 4 <= nchars <= 64 and (not nchars % 4), 'illegal ‘nchars’ value'
		ret = ' '.join(ret[i:i+4] for i in range(0, nchars, 4))
	else:
		assert 4 <= nchars <= 64, 'illegal ‘nchars’ value'
	return ret.upper() if upper else ret

def make_chksum_8(s, *, sep=False):
	from .obj import HexStr
	from hashlib import sha256
	s = HexStr(sha256(sha256(s).digest()).hexdigest()[:8].upper(), case='upper')
	return '{} {}'.format(s[:4], s[4:]) if sep else s

def make_chksum_6(s):
	from .obj import HexStr
	from hashlib import sha256
	if isinstance(s, str):
		s = s.encode()
	return HexStr(sha256(s).hexdigest()[:6])

def is_chksum_6(s):
	return len(s) == 6 and set(s) <= set(hexdigits_lc)

def split_into_cols(col_wid, s):
	return ' '.join([s[col_wid*i:col_wid*(i+1)] for i in range(len(s)//col_wid+1)]).rstrip()

def capfirst(s): # different from str.capitalize() - doesn't downcase any uc in string
	return s if len(s) == 0 else s[0].upper() + s[1:]

def decode_timestamp(s):
#	tz_save = open('/etc/timezone').read().rstrip()
	os.environ['TZ'] = 'UTC'
#	os.environ['TZ'] = tz_save
	return int(time.mktime(time.strptime(s, '%Y%m%d_%H%M%S')))

def make_timestamp(secs=None):
	return '{:04d}{:02d}{:02d}_{:02d}{:02d}{:02d}'.format(*time.gmtime(
		int(secs) if secs is not None else time.time())[:6])

def make_timestr(secs=None):
	return '{}-{:02d}-{:02d} {:02d}:{:02d}:{:02d}'.format(*time.gmtime(
		int(secs) if secs is not None else time.time())[:6])

def secs_to_dhms(secs):
	hrs = secs // 3600
	return '{}{:02d}:{:02d}:{:02d} h/m/s'.format(
		('{} day{}, '.format(hrs//24, suf(hrs//24)) if hrs > 24 else ''),
		hrs % 24,
		(secs // 60) % 60,
		secs % 60
	)

def secs_to_hms(secs):
	return '{:02d}:{:02d}:{:02d}'.format(secs//3600, (secs//60) % 60, secs % 60)

def secs_to_ms(secs):
	return '{:02d}:{:02d}'.format(secs//60, secs % 60)

def is_int(s): # actually is_nonnegative_int()
	return set(str(s) or 'x') <= set(digits)

def check_int_between(val, imin, imax, *, desc):
	if not imin <= int(val) <= imax:
		die(1, f'{val}: invalid value for {desc} (must be between {imin} and {imax})')
	return int(val)

def check_member(e, iterable, desc, message='unsupported'):
	if e not in iterable:
		from mmgen.color import yellow
		die(1, yellow(f'{e}: {message} {desc} (must be one of {fmt_list(iterable)})'))

def is_hex_str(s):
	return set(s) <= set(hexdigits)

def is_hex_str_lc(s):
	return set(s) <= set(hexdigits_lc)

def is_utf8(s):
	try:
		s.decode('utf8')
	except:
		return False
	else:
		return True

def remove_whitespace(s, *, ws='\t\r\n '):
	return s.translate(dict((ord(e), None) for e in ws))

def strip_comment(line):
	return re.sub('#.*', '', line).rstrip()

def strip_comments(lines):
	pat = re.compile('#.*')
	return [m for m in [pat.sub('', l).rstrip() for l in lines] if m != '']

def make_full_path(outdir, outfile):
	return os.path.normpath(os.path.join(outdir, os.path.basename(outfile)))

class oneshot_warning:

	color = 'nocolor'

	def __init__(self, *, div=None, fmt_args=[], reverse=False):
		self.do(type(self), div, fmt_args, reverse)

	def do(self, wcls, div, fmt_args, reverse):

		def do_warning():
			from . import color
			msg(getattr(color, getattr(wcls, 'color'))('WARNING: ' + getattr(wcls, 'message').format(*fmt_args)))

		if not hasattr(wcls, 'data'):
			setattr(wcls, 'data', [])

		data = getattr(wcls, 'data')
		condition = (div in data) if reverse else (not div in data)

		if not div in data:
			data.append(div)

		if condition:
			do_warning()
			self.warning_shown = True
		else:
			self.warning_shown = False

class oneshot_warning_group(oneshot_warning):

	def __init__(self, wcls, *, div=None, fmt_args=[], reverse=False):
		self.do(getattr(self, wcls), div, fmt_args, reverse)

def get_subclasses(cls, *, names=False):
	def gen(cls):
		for i in cls.__subclasses__():
			yield i
			yield from gen(i)
	return tuple((c.__name__ for c in gen(cls)) if names else gen(cls))

def async_run(coro):
	import asyncio
	return asyncio.run(coro)

def wrap_ripemd160(called=[]):
	if not called:
		try:
			import hashlib
			hashlib.new('ripemd160')
		except ValueError:
			def hashlib_new_wrapper(name, *args, **kwargs):
				if name == 'ripemd160':
					return ripemd160(*args, **kwargs)
				else:
					return hashlib_new(name, *args, **kwargs)
			from .contrib.ripemd160 import ripemd160
			hashlib_new = hashlib.new
			hashlib.new = hashlib_new_wrapper
		called.append(True)

def exit_if_mswin(feature):
	if sys.platform == 'win32':
		die(2, capfirst(feature) + ' not supported on the MSWin / MSYS2 platform')

def have_sudo(*, silent=False):
	from subprocess import run, DEVNULL
	redir = DEVNULL if silent else None
	try:
		run(['sudo', '--non-interactive', 'true'], stdout=redir, stderr=redir, check=True)
		return True
	except:
		return False

def in_nix_environment():
	for path in os.getenv('PATH').split(':'):
		if path.startswith('/nix/store/'):
			return True

def cached_property(orig_func):
	@property
	def new_func(self):
		attr_name = '_' + orig_func.__name__
		if not hasattr(self, attr_name):
			setattr(self, attr_name, orig_func(self))
		return getattr(self, attr_name)
	return new_func



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.btc.tx.new_swap: Bitcoin new swap transaction class
"""

from ....tx.new_swap import NewSwap as TxNewSwap
from .new import New

class NewSwap(New, TxNewSwap):
	desc = 'Bitcoin swap transaction'

	def update_data_output(self, trade_limit):
		o = self.data_output._asdict()
		parsed_memo = self.swap_proto_mod.Memo.parse(o['data'].decode())
		memo = self.swap_proto_mod.Memo(
			self.swap_cfg,
			self.recv_proto,
			self.recv_asset,
			self.recv_proto.coin_addr(parsed_memo.address),
			trade_limit = trade_limit)
		o['data'] = f'data:{memo}'
		self.data_output = self.Output(self.proto, **o)

	@property
	def vault_idx(self):
		return self._chg_output_ops('idx', 'is_vault')

	@property
	def vault_output(self):
		return self._chg_output_ops('output', 'is_vault')



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
altcoin.util: various altcoin-related utilities
"""

from ..util import die

def decrypt_keystore(data, passwd, *, mac_algo=None, mac_params={}):
	"""
	Decrypt the encrypted data in a cross-chain keystore
	Returns the decrypted data as a bytestring
	"""

	cdata = data['crypto']
	parms = cdata['kdfparams']

	valid_kdfs = ['scrypt', 'pbkdf2']
	valid_ciphers = ['aes-128-ctr', 'aes-256-ctr']

	if (kdf := cdata['kdf']) not in valid_kdfs:
		die(1, f'unsupported key derivation function {kdf!r} (must be one of {valid_kdfs})')

	if (cipher := cdata['cipher']) not in valid_ciphers:
		die(1, f'unsupported cipher {cipher!r} (must be one of {valid_ciphers})')

	# Derive encryption key from password:
	if kdf == 'scrypt':
		if not mac_algo:
			die(1, 'the ‘mac_algo’ parameter is required for scrypt kdf')
		from hashlib import scrypt
		hashed_pw = scrypt(
			password = passwd,
			salt     = bytes.fromhex(parms['salt']),
			n        = parms['n'],
			r        = parms['r'],
			p        = parms['p'],
			maxmem   = 0,
			dklen    = parms['dklen'])
	elif kdf == 'pbkdf2':
		if (prf := parms.get('prf')) != 'hmac-sha256':
			die(1, f"unsupported hash function {prf!r} (must be 'hmac-sha256')")
		from cryptography.hazmat.primitives import hashes
		from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
		hashed_pw = PBKDF2HMAC(
			algorithm  = hashes.SHA256(),
			length     = parms['dklen'],
			salt       = bytes.fromhex(parms['salt']),
			iterations = parms['c']
		).derive(passwd)
		# see:
		#   https://github.com/xchainjs/xchainjs-lib.git
		#   https://github.com/xchainjs/foundry-primitives-js.git
		from hashlib import blake2b
		mac_algo = mac_algo or blake2b
		mac_params = mac_params or {'digest_size': 32}

	mac = mac_algo(
		hashed_pw[16:32] + bytes.fromhex(cdata['ciphertext']),
		**mac_params
	).digest().hex()

	if mac != cdata['mac']:
		die(1, 'incorrect password')

	# Decrypt data:
	from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
	from cryptography.hazmat.backends import default_backend
	cipher_len = int(cipher.split('-')[1]) // 8
	c = Cipher(
		algorithms.AES(hashed_pw[:cipher_len]),
		modes.CTR(bytes.fromhex(cdata['cipherparams']['iv'])),
		backend = default_backend())
	encryptor = c.encryptor()
	return encryptor.update(bytes.fromhex(cdata['ciphertext'])) + encryptor.finalize()



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
daemon: Daemon control interface for the MMGen suite
"""

import sys, os, time, importlib
from subprocess import run, PIPE, CompletedProcess
from collections import namedtuple

from .base_obj import Lockable
from .color import set_vt100
from .util import msg, Msg_r, die, remove_dups, oneshot_warning, fmt_list
from .flags import ClassFlags, ClassOpts

_dd = namedtuple('daemon_data', ['coind_name', 'coind_version', 'coind_version_str']) # latest tested version
_nw = namedtuple('coin_networks', ['mainnet', 'testnet', 'regtest'])

class Daemon(Lockable):

	desc = 'daemon'
	debug = False
	wait = True
	use_pidfile = True
	force_kill = False
	pids = ()
	use_threads = False
	cfg_file = None
	new_console_mswin = False
	lockfile = None
	private_port = None
	avail_opts = ()
	avail_flags = () # like opts, but can be set or unset after instantiation
	_reset_ok = ('debug', 'wait', 'pids')
	version_info_arg = '--version'

	def __init__(self, cfg, *, opts=None, flags=None):

		self.cfg = cfg
		self.platform = sys.platform
		if self.platform == 'win32':
			self.use_pidfile = False
			self.use_threads = True

		self.opt = ClassOpts(self, opts)
		self.flag = ClassFlags(self, flags)
		self.debug = self.debug or cfg.debug_daemon

	def exec_cmd_thread(self, cmd):
		import threading
		tname = ('exec_cmd', 'exec_cmd_win_console')[self.platform == 'win32' and self.new_console_mswin]
		t = threading.Thread(target=getattr(self, tname), args=(cmd,))
		t.daemon = True
		t.start()
		if self.platform == 'win32':
			Msg_r(' \b') # blocks w/o this...crazy
		return True

	def exec_cmd_win_console(self, cmd):
		from subprocess import Popen, CREATE_NEW_CONSOLE, STARTUPINFO, STARTF_USESHOWWINDOW, SW_HIDE
		si = STARTUPINFO(dwFlags=STARTF_USESHOWWINDOW, wShowWindow=SW_HIDE)
		p = Popen(cmd, creationflags=CREATE_NEW_CONSOLE, startupinfo=si)
		p.wait()

	def exec_cmd(self, cmd, *, is_daemon=False, check_retcode=False):
		out = (PIPE, None)[is_daemon and self.opt.no_daemonize]
		try:
			cp = run(cmd, check=False, stdout=out, stderr=out)
		except OSError as e:
			die('MMGenCalledProcessError', f'Error starting executable: {type(e).__name__} [Errno {e.errno}]')
		set_vt100()
		if check_retcode and cp.returncode:
			die(1, str(cp))
		if self.debug:
			print(cp)
		return cp

	def run_cmd(self, cmd, *, silent=False, is_daemon=False, check_retcode=False):

		if self.debug:
			msg('\n\n')

		if self.debug or (is_daemon and not silent):
			msg(f'Starting {self.desc} on port {self.bind_port}')

		if self.debug:
			msg(f'\nExecuting:\n{fmt_list(cmd, fmt="col", indent="  ")}\n')

		if self.use_threads and is_daemon and not self.opt.no_daemonize:
			ret = self.exec_cmd_thread(cmd)
		else:
			ret = self.exec_cmd(cmd, is_daemon=is_daemon, check_retcode=check_retcode)

		if isinstance(ret, CompletedProcess):
			if ret.stdout and (self.debug or not silent):
				msg(ret.stdout.decode().rstrip())
			if ret.stderr and (self.debug or (ret.returncode and not silent)):
				msg(ret.stderr.decode().rstrip())

		return ret

	@property
	def pid(self):
		if self.use_pidfile:
			with open(self.pidfile) as fp:
				return fp.read().strip()
		elif self.platform == 'win32':
			# Assumes only one running instance of given daemon.  If multiple daemons are running,
			# the first PID in the list is returned and self.pids is set to the PID list.
			ss = f'{self.exec_fn}.exe'
			cp = self.run_cmd(['ps', '-Wl'], silent=True)
			self.pids = ()
			# use Windows, not Cygwin, PID
			pids = tuple(line.split()[3] for line in cp.stdout.decode().splitlines() if ss in line)
			if pids:
				if len(pids) > 1:
					self.pids = pids
				return pids[0]
		elif self.platform in ('linux', 'darwin'):
			ss = ' '.join(self.start_cmd)
			cp = self.run_cmd(['pgrep', '-f', ss], silent=True)
			if cp.stdout:
				return cp.stdout.strip().decode()
		die(2, f'{ss!r} not found in process list, cannot determine PID')

	@property
	def bind_port(self):
		return self.private_port or self.rpc_port

	@property
	def state(self):
		if self.debug:
			msg(f'Testing port {self.bind_port}')
		return 'ready' if self.test_socket('localhost', self.bind_port) else 'stopped'

	@property
	def start_cmds(self):
		return [self.start_cmd]

	@property
	def stop_cmd(self):
		return (
			['kill', '-Wf', self.pid] if self.platform == 'win32' else
			['kill', '-9', self.pid] if self.force_kill else
			['kill', self.pid])

	def cmd(self, action, *args, **kwargs):
		return getattr(self, action)(*args, **kwargs)

	def cli(self, *cmds, silent=False):
		return self.run_cmd(self.cli_cmd(*cmds), silent=silent)

	def state_msg(self, *, extra_text=None):
		try:
			pid = self.pid
		except:
			pid = None
		extra_text = 'not ' if self.state == 'stopped' else f'{extra_text} ' if extra_text else ''
		return '{:{w}} {:10} {}'.format(
			f'{self.desc} {extra_text}running',
			'pid N/A' if pid is None or self.pids or self.state == 'stopped' else f'pid {pid}',
			f'port {self.bind_port}',
			w = 60)

	def pre_start(self):
		pass

	def start(self, *, quiet=False, silent=False):
		if self.state == 'ready':
			if not (quiet or silent):
				msg(self.state_msg(extra_text='already'))
			return True

		self.wait_for_state('stopped')

		self.pre_start()

		if not silent:
			msg(f'Starting {self.desc} on port {self.bind_port}')
		ret = self.run_cmd(self.start_cmd, silent=True, is_daemon=True, check_retcode=True)

		if self.wait:
			self.wait_for_state('ready')

		return ret

	def stop(self, *, quiet=False, silent=False):
		if self.state == 'ready':
			if not silent:
				msg(f'Stopping {self.desc} on port {self.bind_port}')
			if self.force_kill:
				run(['sync'])
			ret = self.run_cmd(self.stop_cmd, silent=True)

			if self.pids:
				msg('Warning: multiple PIDs [{}] -- we may be stopping the wrong instance'.format(
					fmt_list(self.pids, fmt='bare')
				))
			if self.wait:
				self.wait_for_state('stopped')
				time.sleep(0.3) # race condition
			return ret
		else:
			if not (quiet or silent):
				msg(f'{self.desc} on port {self.bind_port} not running')
			return True

	def restart(self, *, silent=False):
		self.stop(silent=silent)
		return self.start(silent=silent)

	def test_socket(self, host, port, *, timeout=10):
		import socket
		try:
			socket.create_connection((host, port), timeout=timeout).close()
		except:
			return False
		else:
			return True

	def wait_for_state(self, req_state):
		for _ in range(self.cfg.daemon_state_timeout * 5):
			if self.state == req_state:
				return True
			time.sleep(0.2)
		die(2, f'Wait for state {req_state!r} timeout exceeded for {self.desc} (port {self.bind_port})')

	@classmethod
	def get_exec_version_str(cls):
		try:
			cp = run([cls.exec_fn, cls.version_info_arg], stdout=PIPE, stderr=PIPE, check=True)
		except Exception as e:
			die(2, f'{e}\nUnable to execute {cls.exec_fn}')

		if cp.returncode:
			die(2, f'Unable to execute {cls.exec_fn}')
		else:
			res = cp.stdout.decode().splitlines()
			return (res[0] if len(res) == 1 else [s for s in res if 'ersion' in s][0]).strip()

class RPCDaemon(Daemon):

	avail_opts = ('no_daemonize',)

	def __init__(self, cfg, *, opts=None, flags=None):
		super().__init__(cfg, opts=opts, flags=flags)
		self.desc = '{} {} {}RPC daemon'.format(
			self.rpc_desc,
			getattr(self.proto.network_names, self.proto.network),
			'test suite ' if self.test_suite else '')
		self._set_ok += ('usr_daemon_args',)
		self.usr_daemon_args = []

	@property
	def start_cmd(self):
		return [self.exec_fn] + self.daemon_args + self.usr_daemon_args

class CoinDaemon(Daemon):
	networks = ('mainnet', 'testnet', 'regtest')
	cfg_file_hdr = ''
	avail_flags = ('keep_cfg_file',)
	avail_opts = ('no_daemonize', 'online')
	testnet_dir = None
	test_suite_port_shift = 1237
	rpc_user = None
	rpc_password = None

	_cd = namedtuple('coins_data', ['daemon_ids'])
	coins = {
		'BTC': _cd(['bitcoin_core']),
		'BCH': _cd(['bitcoin_cash_node']),
		'LTC': _cd(['litecoin_core']),
		'XMR': _cd(['monero']),
		'ETH': _cd(['geth', 'reth', 'erigon']), #, 'openethereum'
		'ETC': _cd(['parity']),
	}

	@classmethod
	def all_daemon_ids(cls):
		return [i for coin in cls.coins for i in cls.coins[coin].daemon_ids]

	class warn_blacklisted(oneshot_warning):
		color = 'yellow'
		message = 'blacklisted daemon: {!r}'

	@classmethod
	def get_daemon_ids(cls, cfg, coin):

		ret = cls.coins[coin].daemon_ids
		if 'erigon' in ret and not cfg.enable_erigon:
			ret.remove('erigon')
		if cfg.blacklisted_daemons:
			blacklist = cfg.blacklisted_daemons.split()
			def gen():
				for daemon_id in ret:
					if daemon_id in blacklist:
						cls.warn_blacklisted(div=daemon_id, fmt_args=[daemon_id])
					else:
						yield daemon_id
			ret = list(gen())
		return ret

	@classmethod
	def get_daemon(cls, cfg, coin, daemon_id, *, proto=None):
		if proto:
			proto_cls = type(proto)
		else:
			from .protocol import init_proto
			proto_cls = init_proto(cfg, coin, return_cls=True)
		return getattr(
			importlib.import_module(f'mmgen.proto.{proto_cls.base_proto_coin.lower()}.daemon'),
			daemon_id+'_daemon')

	@classmethod
	def get_network_ids(cls, cfg):
		from .protocol import CoinProtocol
		def gen():
			for coin in cls.coins:
				for daemon_id in cls.get_daemon_ids(cfg, coin):
					for network in cls.get_daemon(cfg, coin, daemon_id).networks:
						yield CoinProtocol.Base.create_network_id(coin, network)
		return remove_dups(list(gen()), quiet=True)

	def __new__(cls,
			cfg,
			*,
			network_id = None,
			proto      = None,
			opts       = None,
			flags      = None,
			test_suite = False,
			port_shift = None,
			p2p_port   = None,
			datadir    = None,
			daemon_id  = None):

		assert network_id or proto,        'CoinDaemon_chk1'
		assert not (network_id and proto), 'CoinDaemon_chk2'

		if proto:
			network_id = proto.network_id
			network    = proto.network
			coin       = proto.coin
		else:
			network_id = network_id.lower()
			from .protocol import CoinProtocol, init_proto
			proto = init_proto(cfg, network_id=network_id)
			coin, network = CoinProtocol.Base.parse_network_id(network_id)
			coin = coin.upper()

		daemon_ids = cls.get_daemon_ids(cfg, coin)
		if not daemon_ids:
			die(1, f'No configured daemons for coin {coin}!')
		daemon_id = (
			daemon_id
			or getattr(cfg, f'{coin.lower()}_daemon_id', None)
			or cfg.daemon_id
			or daemon_ids[0])

		if daemon_id not in daemon_ids:
			die(1, f'{daemon_id!r}: invalid daemon_id - valid choices: {fmt_list(daemon_ids)}')

		me = Daemon.__new__(cls.get_daemon(cfg, None, daemon_id, proto=proto))

		assert network in me.networks, f'{network!r}: unsupported network for daemon {daemon_id}'
		me.network_id = network_id
		me.network = network
		me.coin = coin
		me.id = daemon_id
		me.proto = proto

		return me

	def __init__(self,
			cfg,
			*,
			network_id = None,
			proto      = None,
			opts       = None,
			flags      = None,
			test_suite = False,
			port_shift = None,
			p2p_port   = None,
			datadir    = None,
			daemon_id  = None):

		self.test_suite = test_suite

		super().__init__(cfg=cfg, opts=opts, flags=flags)

		self._set_ok += ('shared_args', 'usr_coind_args')
		self.shared_args = []
		self.usr_coind_args = []

		for k, v in self.daemon_data._asdict().items():
			setattr(self, k, v)

		self.desc = '{} {} {}daemon'.format(
			self.coind_name,
			getattr(self.proto.network_names, self.network),
			'test suite ' if test_suite else '')

		# user-set values take precedence
		self.datadir = os.path.abspath(datadir or cfg.daemon_data_dir or self.init_datadir())
		self.non_dfl_datadir = bool(datadir or cfg.daemon_data_dir or test_suite or self.network == 'regtest')

		# init_datadir() may have already initialized logdir
		self.logdir = os.path.abspath(getattr(self, 'logdir', self.datadir))

		ps_adj = (port_shift or 0) + (self.test_suite_port_shift if test_suite else 0)

		# user-set values take precedence
		usr_rpc_port = self.proto.rpc_port or cfg.rpc_port
		self.rpc_port = usr_rpc_port + (port_shift or 0) if usr_rpc_port else ps_adj + self.get_rpc_port()
		self.p2p_port = (
			p2p_port or (
				self.get_p2p_port() + ps_adj if self.get_p2p_port() and (test_suite or ps_adj) else None
			) if self.network != 'regtest' else None)

		if hasattr(self, 'private_ports'):
			self.private_port = getattr(self.private_ports, self.network)

		# bind_port == self.private_port or self.rpc_port
		self.pidfile = f'{self.logdir}/{self.id}-{self.network}-daemon-{self.bind_port}.pid'
		self.logfile = f'{self.logdir}/{self.id}-{self.network}-daemon-{self.bind_port}.log'

		self.init_subclass()

	def init_datadir(self):
		if self.test_suite:
			return os.path.join('test', 'daemons', self.network_id)
		else:
			return os.path.join(*self.datadirs[self.platform])

	@property
	def network_datadir(self):
		return self.datadir

	def get_rpc_port(self):
		return getattr(self.rpc_ports, self.network)

	def get_p2p_port(self):
		return None

	@property
	def start_cmd(self):
		return ([self.exec_fn]
				+ self.coind_args
				+ self.shared_args
				+ self.usr_coind_args)

	def cli_cmd(self, *cmds):
		return ([self.cli_fn]
				+ self.shared_args
				+ list(cmds))

	def start(self, *args, **kwargs):
		assert self.test_suite or self.network == 'regtest', 'start() restricted to test suite and regtest'
		return super().start(*args, **kwargs)

	def stop(self, *args, **kwargs):
		assert self.test_suite or self.network == 'regtest', 'stop() restricted to test suite and regtest'
		return super().stop(*args, **kwargs)

	def pre_start(self):
		os.makedirs(self.datadir, exist_ok=True)

		if self.test_suite or self.network == 'regtest':
			if self.cfg_file and not self.flag.keep_cfg_file:
				with open(f'{self.datadir}/{self.cfg_file}', 'w') as fp:
					fp.write(self.cfg_file_hdr)

		if self.use_pidfile and os.path.exists(self.pidfile):
			# Parity overwrites the data in the existing pidfile without zeroing it first, leading
			# to interesting consequences when the new PID has fewer digits than the previous one.
			os.unlink(self.pidfile)

	def remove_datadir(self):
		"remove the network's datadir"
		assert self.test_suite, 'datadir removal restricted to test suite'
		if self.state == 'stopped':
			run(['rm', '-rf', self.datadir])
			set_vt100()
		else:
			msg(f'Cannot remove {self.network_datadir!r} - daemon is not stopped')



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
test/hashfunc.py: Test internal implementations of SHA256, SHA512 and Keccak256
"""

import sys

try:
	from include import test_init
except ImportError:
	from test.include import test_init

from mmgen.util import msg, msg_r, die

def green(s):
	return '\033[32;1m' + s + '\033[0m'

class TestHashFunc:

	def test_constants(self):
		msg_r('Testing generated constants: ')
		h = self.t_cls(b'foo')
		if h.H_init != self.H_ref:
			m = 'Generated constants H[] differ from reference value:\nReference:\n{}\nGenerated:\n{}'
			die(3, m.format([hex(n) for n in self.H_ref], [hex(n) for n in h.H_init]))
		if h.K != self.K_ref:
			m = 'Generated constants K[] differ from reference value:\nReference:\n{}\nGenerated:\n{}'
			die(3, m.format([hex(n) for n in self.K_ref], [hex(n) for n in h.K]))
		msg('OK')

	def compare_hashes(self, data, chk=None):
		if chk is None:
			chk = getattr(self.hashlib, self.desc)(data).hexdigest()
		res = self.t_cls(data).hexdigest()
		if res != chk:
			m ='\nHashes do not match!\nReference {d}: {}\nMMGen {d}:     {}'
			die(3, m.format(chk, res, d=self.desc.upper()))

	def test_ref(self):
		for i, data in enumerate(self.vectors):
			msg_r(f'\rTesting reference input data: {i+1:4}/{len(self.vectors)} ')
			self.compare_hashes(data.encode(), chk=self.vectors[data])
		msg('OK')

	def test_random(self, rounds):
		if not self.hashlib:
			return
		for i in range(rounds):
			if i+1 in (1, rounds) or not (i+1) % 10:
				msg_r(f'\rTesting random input data:    {i+1:4}/{rounds} ')
			dlen = int(getrand(4).hex(), 16) >> 18
			self.compare_hashes(getrand(dlen))
		msg('OK')

class TestKeccak(TestHashFunc):
	desc = 'keccak_256'
	vectors = {
		'':                           'c5d2460186f7233c927e7db2dcc703c0e500b653ca82273b7bfad8045d85a470',
		'x':                          '7521d1cadbcfa91eec65aa16715b94ffc1c9654ba57ea2ef1a2127bca1127a83',
		'xa':                         'ac3f7abb3970ce4fd85fbf46ddd6c87f63e9f91deb38aa155c237c4f81a74259',
		'the':                        '58d5df6c336f348e541c83745572ac73656a0238a55c006a84123e2ace2e7aef',
		'7_chars':                    '537d95bb6222fc8d333177a4310d5549c33aae32c638a01f8ca13f8271d49a89',
		'8charmsg':                   '590d8871128ec3b84d833e2d36cc7e9e1b808b88d44aae653d6938f7e23af707',
		'9_charmsg':                  '625b2ad4132fb699ba33a8cb7c8f1a11f656054cd2fac2156ddd92585c26caac',
		'10_charmsg':                 'b24434a23e05b2f2f1a575df7647d1e6cb96e9c730190f8bc6b282aa18ff96fb',
		'8charmsg' * 8:               'b90078637f407bde1b4e44bef157810b596430a375a4f094ab3fec5258aafe8f',
		'8charmsg' * 7 + '7_chars':   'd566a70e4db4592d2649289024f6deaa1fff7138591e27c1844d172a9d16f2fe',
		'8charmsg' * 7 + '9_charmsg': '149a1620df1447401c968f9593b190dbd519ed663cde7f8606c75d6121e7c980',
		'\x00':                       'bc36789e7a1e281436464229828f817d6612f7b477d66591ff96a9e064bcc98a',
		'\x00\x00':                   '54a8c0ab653c15bfb48b47fd011ba2b9617af01cb45cab344acd57c924d56798',
		'\x00'*256:                   'd397b3b043d87fcd6fad1291ff0bfd16401c274896d8c63a923727f077b8e0b5',
		'\x0f':                       '3d725c5ee53025f027da36bea8d3af3b6a3e9d2d1542d47c162631de48e66c1c',
		'\x0f\x0f':                   'c5c1cadbf3ce871e667034a04649cafecd77792491ca7d75207bff6834725e5c',
		'\x0f'*256:                   '634cd5b0c7131441a789340d2ddb9158809bcee04e3389428a99094c586ab9e3',
		'\x0f\x0d':                   'ce69f143c196075c65560b98431330ed46e3bc14daeda0ed1dd7124be89f6c16',
		'\x0e\x0e'*256:               'ccd35755ad4c2bf5dd30b8867e1bb956fceebb6ec95e3684362f1ce160527481',
		'\x00'*511:                   'd28fcca01972cfdea3de8a981db12e426a4ffa25830c25f213e4823c41ab90f3',
		'\x00'*512:                   'd5c44f659751a819616c58c9efe38e80f2b84cf621036da99c019bbe4f1fb647',
		'\x00'*513:                   '8a1810ea3d9e5cc95089994be993c99382f2a7754e8c94ee14888b56ecaa05fd',
		'\x0f'*512:                   'eae2f92f9d3564b346a1937626ca25b9fafb19433a624008c0b6c0eec469aa7c',
		'\x0f'*511:                   'da69b07f366d33f62eb889e7bf05705dae4aaf8902bb8dece86346d33b540e71',
		'\x0f'*513:                   '131640bf7001195b269d652efe137afdc17f19b403b4b5f9aba09a6fb8cbecbe',
		'\x00\x0f'*512:               '330bd984f6c6361b02d7a5de6d4fc60ae8e72c7b791a511d428090b0c9af2313',
		'\x0e\x0f'*511:               '5d17f4e0f63342d3d635bfe57c416a4dddc76ccb5bb6981aacdf3227d095d244',
		'\x0a\x0d'*513:               '187c8fa19c9ac5dedbe4a3078290120eb2f5f7b3b0cb4590cdfbebcfb37bf2b3',
		'\x00\x0f'*1024:              '7aa8a949809ac159cb98fe947ddc40de05be9309aff563130d7d55afb4cc9e11',
		'\x0e\x0f'*1023:              '5c82fa38e3d35efe65a392835be500ba62d0dfc9c0de28950cdeb5bc0d96689d',
		'\x0a\x0d'*1025:              'bc99495fcaf4a987c421ff1cc5854b25786964ab863836b34c82a437a473f4c4',
	}

	def __init__(self):
		from mmgen.contrib.keccak import keccak_256
		from mmgen.util2 import get_keccak
		self.t_cls = keccak_256
		class hashlib:
			keccak_256 = get_keccak()
		self.hashlib = hashlib

	def test_constants(self):
		pass

class TestSha2(TestHashFunc):
	desc = 'sha2'

	def __init__(self):
		from mmgen.sha2 import Sha256, Sha512
		import hashlib
		self.t_cls = {'sha256':Sha256, 'sha512':Sha512}[self.desc]
		self.hashlib = hashlib
		self.vectors = {k:None for k in TestKeccak.vectors}

class TestSha256(TestSha2):
	desc = 'sha256'
	H_ref = (
		0x6a09e667, 0xbb67ae85, 0x3c6ef372, 0xa54ff53a, 0x510e527f, 0x9b05688c, 0x1f83d9ab, 0x5be0cd19)
	K_ref = (
		0x428a2f98, 0x71374491, 0xb5c0fbcf, 0xe9b5dba5, 0x3956c25b, 0x59f111f1, 0x923f82a4, 0xab1c5ed5,
		0xd807aa98, 0x12835b01, 0x243185be, 0x550c7dc3, 0x72be5d74, 0x80deb1fe, 0x9bdc06a7, 0xc19bf174,
		0xe49b69c1, 0xefbe4786, 0x0fc19dc6, 0x240ca1cc, 0x2de92c6f, 0x4a7484aa, 0x5cb0a9dc, 0x76f988da,
		0x983e5152, 0xa831c66d, 0xb00327c8, 0xbf597fc7, 0xc6e00bf3, 0xd5a79147, 0x06ca6351, 0x14292967,
		0x27b70a85, 0x2e1b2138, 0x4d2c6dfc, 0x53380d13, 0x650a7354, 0x766a0abb, 0x81c2c92e, 0x92722c85,
		0xa2bfe8a1, 0xa81a664b, 0xc24b8b70, 0xc76c51a3, 0xd192e819, 0xd6990624, 0xf40e3585, 0x106aa070,
		0x19a4c116, 0x1e376c08, 0x2748774c, 0x34b0bcb5, 0x391c0cb3, 0x4ed8aa4a, 0x5b9cca4f, 0x682e6ff3,
		0x748f82ee, 0x78a5636f, 0x84c87814, 0x8cc70208, 0x90befffa, 0xa4506ceb, 0xbef9a3f7, 0xc67178f2)

class TestSha512(TestSha2):
	desc = 'sha512'
	H_ref = (
		0x6a09e667f3bcc908, 0xbb67ae8584caa73b, 0x3c6ef372fe94f82b, 0xa54ff53a5f1d36f1,
		0x510e527fade682d1, 0x9b05688c2b3e6c1f, 0x1f83d9abfb41bd6b, 0x5be0cd19137e2179)
	K_ref = (
		0x428a2f98d728ae22, 0x7137449123ef65cd, 0xb5c0fbcfec4d3b2f, 0xe9b5dba58189dbbc, 0x3956c25bf348b538,
		0x59f111f1b605d019, 0x923f82a4af194f9b, 0xab1c5ed5da6d8118, 0xd807aa98a3030242, 0x12835b0145706fbe,
		0x243185be4ee4b28c, 0x550c7dc3d5ffb4e2, 0x72be5d74f27b896f, 0x80deb1fe3b1696b1, 0x9bdc06a725c71235,
		0xc19bf174cf692694, 0xe49b69c19ef14ad2, 0xefbe4786384f25e3, 0x0fc19dc68b8cd5b5, 0x240ca1cc77ac9c65,
		0x2de92c6f592b0275, 0x4a7484aa6ea6e483, 0x5cb0a9dcbd41fbd4, 0x76f988da831153b5, 0x983e5152ee66dfab,
		0xa831c66d2db43210, 0xb00327c898fb213f, 0xbf597fc7beef0ee4, 0xc6e00bf33da88fc2, 0xd5a79147930aa725,
		0x06ca6351e003826f, 0x142929670a0e6e70, 0x27b70a8546d22ffc, 0x2e1b21385c26c926, 0x4d2c6dfc5ac42aed,
		0x53380d139d95b3df, 0x650a73548baf63de, 0x766a0abb3c77b2a8, 0x81c2c92e47edaee6, 0x92722c851482353b,
		0xa2bfe8a14cf10364, 0xa81a664bbc423001, 0xc24b8b70d0f89791, 0xc76c51a30654be30, 0xd192e819d6ef5218,
		0xd69906245565a910, 0xf40e35855771202a, 0x106aa07032bbd1b8, 0x19a4c116b8d2d0c8, 0x1e376c085141ab53,
		0x2748774cdf8eeb99, 0x34b0bcb5e19b48a8, 0x391c0cb3c5c95a63, 0x4ed8aa4ae3418acb, 0x5b9cca4f7763e373,
		0x682e6ff3d6b2b8a3, 0x748f82ee5defb2fc, 0x78a5636f43172f60, 0x84c87814a1f0ab72, 0x8cc702081a6439ec,
		0x90befffa23631e28, 0xa4506cebde82bde9, 0xbef9a3f7b2c67915, 0xc67178f2e372532b, 0xca273eceea26619c,
		0xd186b8c721c0c207, 0xeada7dd6cde0eb1e, 0xf57d4f7fee6ed178, 0x06f067aa72176fba, 0x0a637dc5a2c898a6,
		0x113f9804bef90dae, 0x1b710b35131c471b, 0x28db77f523047d84, 0x32caab7b40c72493, 0x3c9ebe0a15c9bebc,
		0x431d67c49c100d4c, 0x4cc5d4becb3e42b6, 0x597f299cfc657e2a, 0x5fcb6fab3ad6faec, 0x6c44198c4a475817)

from test.include.common import getrand, set_globals
from mmgen.cfg import Config
from mmgen.main import launch

def main():

	if len(sys.argv) not in (2, 3):
		die(1, 'Test takes 1 or 2 arguments: test name, plus optional rounds count')

	test = sys.argv[1].capitalize()

	if test not in ('Sha256', 'Sha512', 'Keccak'):
		die(1, "Valid choices for test are 'sha256', 'sha512' or 'keccak'")

	random_rounds = int(sys.argv[2]) if len(sys.argv) == 3 else 500

	set_globals(Config())
	t = globals()['Test'+test]()
	msg(f'Testing internal implementation of {t.desc}')
	t.test_constants()
	t.test_ref()
	t.test_random(random_rounds)

launch(func=main)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
help: help notes for MMGen suite commands
"""

import sys

from ..cfg import gc

def version(cfg):
	from ..util import fmt
	print(fmt(f"""
		{gc.prog_name.upper()} version {gc.version}
		Part of {gc.proj_name} Wallet, an online/offline cryptocurrency wallet for the
		command line. Copyright (C){gc.Cdates} {gc.author} {gc.email}
	""", indent='  ').rstrip())
	sys.exit(0)

def list_daemon_ids(cfg):
	from ..daemon import CoinDaemon
	from ..util import msg, fmt_list
	msg('  {} {}'.format('Coin', 'Daemon IDs'))
	msg('  {} {}'.format('----', '----------'))
	for k, v in CoinDaemon.coins.items():
		msg('  {}  {}'.format(k, fmt_list(v.daemon_ids, fmt='barest')))
	sys.exit(0)

def show_hash_presets(cfg):
	fs = '      {:<6} {:<3} {:<2} {}'
	from ..util import msg
	from ..crypto import Crypto
	msg('  Available parameters for scrypt.hash():')
	msg(fs.format('Preset', 'N', 'r', 'p'))
	for i in sorted(Crypto.hash_presets.keys()):
		msg(fs.format(i, *Crypto.hash_presets[i]))
	msg('  N = memory usage (power of two)\n  p = iterations (rounds)')
	sys.exit(0)

def gen_arg_tuple(cfg, func, text):

	def help_notes(k, *args, **kwargs):
		import importlib
		return getattr(importlib.import_module(
			f'{cfg._help_pkg}.help_notes').help_notes(proto, cfg), k)(*args, **kwargs)

	def help_mod(modname):
		import importlib
		return importlib.import_module(
			f'{cfg._opts.help_pkg}.{modname}').help(proto, cfg)

	from ..protocol import init_proto_from_cfg
	proto = init_proto_from_cfg(cfg, need_amt=True)

	d = {
		'proto':      proto,
		'help_notes': help_notes,
		'help_mod':   help_mod,
		'cfg':        cfg,
	}

	for arg in func.__code__.co_varnames:
		yield d[arg] if arg in d else text

def make_usage_str(cfg, caller):
	indent, col1_w = {
		'help': (2, len(gc.prog_name) + 1),
		'user': (0, len('USAGE:')),
	}[caller]
	def gen():
		ulbl = 'USAGE:'
		for line in [cfg._usage_data.strip()] if isinstance(cfg._usage_data, str) else cfg._usage_data:
			yield '{a:{w}} {b} {c}'.format(
				a = ulbl,
				b = gc.prog_name,
				c = cfg._usage_code(*gen_arg_tuple(cfg, cfg._usage_code, line)) if cfg._usage_code else line,
				w = col1_w)
			ulbl = ''
	return ('\n' + (' ' * indent)).join(gen())

def usage(cfg):
	print(make_usage_str(cfg, caller='user'))
	sys.exit(0)

class Help:

	def make(self, cfg, opts):

		def gen_output():
			yield '  {} {}'.format(gc.prog_name.upper() + ':', opts.opts_data['text']['desc'].strip())
			yield make_usage_str(cfg, caller='help')
			yield self.help_type.upper().replace('_', ' ') + ':'

			# process code for options
			opts_text = nl.join(self.gen_text(opts))
			if 'options' in code:
				yield code['options'](*gen_arg_tuple(cfg, code['options'], opts_text))
			else:
				yield opts_text

			# process code for notes
			if 'notes' in text:
				if 'notes' in code:
					yield from code['notes'](*gen_arg_tuple(cfg, code['notes'], text['notes'])).splitlines()
				else:
					yield from text['notes'].splitlines()

		text = getattr(opts, self.data_desc)['text']
		code = getattr(opts, self.data_desc).get('code', {})
		nl = '\n  '

		return nl.join(gen_output()) + '\n'

class CmdHelp_v1(Help):

	help_type = 'options'
	data_desc = 'opts_data'

	def gen_text(self, opts):
		from ..opts import cmd_opts_v1_pat
		skipping = False
		for line in opts.opts_data['text']['options'].strip().splitlines():
			if m := cmd_opts_v1_pat.match(line):
				yield '{} --{} {}'.format(
					(f'-{m[1]},', '   ')[m[1] == '-'],
					m[2],
					m[4])
			elif not skipping:
				yield line

class CmdHelp_v2(CmdHelp_v1):

	def gen_text(self, opts):
		from ..opts import cmd_opts_v2_help_pat
		skipping = False
		coin_codes = opts.global_filter_codes.coin
		cmd_codes = opts.opts_data['filter_codes']
		for line in opts.opts_data['text']['options'][1:].rstrip().splitlines():
			m = cmd_opts_v2_help_pat.match(line)
			if m[1] == '+':
				if not skipping:
					yield line[6:]
			elif (coin_codes is None or m[1] in coin_codes) and m[2] in cmd_codes:
				yield '{} --{} {}'.format(
					(f'-{m[3]},', '   ')[m[3] == '-'],
					m[4],
					m[6]
				) if m[4] else m[6]
				skipping = False
			else:
				skipping = True

class GlobalHelp(Help):

	help_type = 'global_options'
	data_desc = 'global_opts_data'

	def gen_text(self, opts):
		from ..opts import global_opts_help_pat
		skipping = False
		coin_codes = opts.global_filter_codes.coin
		cmd_codes = opts.global_filter_codes.cmd
		for line in opts.global_opts_data['text']['options'][1:].rstrip().splitlines():
			m = global_opts_help_pat.match(line)
			if m[1] == '+':
				if not skipping:
					yield line[4:]
			elif (coin_codes is None or m[1] in coin_codes) and (cmd_codes is None or m[2] in cmd_codes):
				yield '  --{} {}'.format(m[3], m[5]) if m[3] else m[5]
				skipping = False
			else:
				skipping = True

def print_help(cfg, opts):

	if cfg.help:
		help_cls = CmdHelp_v2 if 'filter_codes' in opts.opts_data else CmdHelp_v1
	else:
		help_cls = GlobalHelp

	from ..ui import do_pager
	do_pager(help_cls().make(cfg, opts))
	sys.exit(0)
